# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
The acquisition process aims to find the optimal point(s) at which to next evaluate the objective
function, with the aim of minimising it. The functionality in this package implements that process.
It typically uses the current observations of the objective function, or a posterior over those
observations.

In this library, the acquisition rule is the central object of the API, while acquisition functions
are supplementary. This is because some acquisition rules, such as Thompson sampling,
do not require an acquisition function. This contrasts with other libraries which may consider
the acquisition function as the central component of this process and assume Efficient Global
Optimization (EGO) for the acquisition rule.

This package contains acquisition rules, as implementations of
:class:`~trieste.acquisition.rule.AcquisitionRule`, and acquisition functions. It also contains
:class:`AcquisitionFunctionBuilder`\ s which provide a common interface for the rules to build
acquisition functions.

Acquisition rules in this library that use acquisition functions, such as
:class:`EfficientGlobalOptimization`, *maximize* these functions. This defines the sign the
acquisition function should take. Additionally, acquisition functions and builders in this library
are designed to minimize the objective function. For example, we do not provide an implementation of
UCB.
"""
from . import optimizer, rule
from .combination import Product, Reducer, Sum
from .function import (
    GIBBON,
    HIPPO,
    AugmentedExpectedImprovement,
    BatchExpectedImprovement,
    BatchMonteCarloExpectedHypervolumeImprovement,
    BatchMonteCarloExpectedImprovement,
    BayesianActiveLearningByDisagreement,
    ExpectedConstrainedHypervolumeImprovement,
    ExpectedConstrainedImprovement,
    ExpectedFeasibility,
    ExpectedHypervolumeImprovement,
    ExpectedImprovement,
    Fantasizer,
    FastConstraintsFeasibility,
    GreedyContinuousThompsonSampling,
    IntegratedVarianceReduction,
    LocalPenalization,
    MakePositive,
    MinValueEntropySearch,
    MonteCarloAugmentedExpectedImprovement,
    MonteCarloExpectedImprovement,
    MultipleOptimismNegativeLowerConfidenceBound,
    NegativeLowerConfidenceBound,
    NegativePredictiveMean,
    ParallelContinuousThompsonSampling,
    PredictiveVariance,
    ProbabilityOfFeasibility,
    augmented_expected_improvement,
    batch_ehvi,
    batch_expected_improvement,
    bayesian_active_learning_by_disagreement,
    bichon_ranjan_criterion,
    expected_hv_improvement,
    expected_improvement,
    fast_constraints_feasibility,
    gibbon_quality_term,
    gibbon_repulsion_term,
    hard_local_penalizer,
    integrated_variance_reduction,
    lower_confidence_bound,
    min_value_entropy_search,
    multiple_optimism_lower_confidence_bound,
    predictive_variance,
    probability_below_threshold,
    soft_local_penalizer,
)
from .interface import (
    AcquisitionFunction,
    AcquisitionFunctionBuilder,
    AcquisitionFunctionClass,
    GreedyAcquisitionFunctionBuilder,
    PenalizationFunction,
    SingleModelAcquisitionBuilder,
    SingleModelGreedyAcquisitionBuilder,
    SingleModelVectorizedAcquisitionBuilder,
    UpdatablePenalizationFunction,
    VectorizedAcquisitionFunctionBuilder,
)
from .rule import (
    AcquisitionRule,
    AsynchronousGreedy,
    AsynchronousOptimization,
    BatchHypervolumeSharpeRatioIndicator,
    BatchTrustRegion,
    BatchTrustRegionBox,
    BatchTrustRegionProduct,
    DiscreteThompsonSampling,
    EfficientGlobalOptimization,
    FixedPointTrustRegionDiscrete,
    HypercubeTrustRegion,
    LocalDatasetsAcquisitionRule,
    RandomSampling,
    SingleObjectiveTrustRegionBox,
    SingleObjectiveTrustRegionDiscrete,
    TREGOBox,
    TURBOBox,
    UpdatableSearchSpace,
    UpdatableTrustRegion,
    UpdatableTrustRegionBox,
    UpdatableTrustRegionDiscrete,
    UpdatableTrustRegionProduct,
)
from .sampler import (
    ExactThompsonSampler,
    GumbelSampler,
    ThompsonSampler,
    ThompsonSamplerFromTrajectory,
)



# Copyright 2022 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Experimental plotting package. Not intended for production code, as it is not yet fully
tested and may change quickly.
"""

try:
    from .inequality_constraints import (
        plot_2obj_cst_query_points,
        plot_init_query_points,
        plot_objective_and_constraints,
    )
    from .plotting import (
        convert_figure_to_frame,
        convert_frames_to_gif,
        plot_acq_function_2d,
        plot_bo_points,
        plot_function_2d,
        plot_gp_2d,
        plot_mobo_history,
        plot_mobo_points_in_obj_space,
        plot_regret,
        plot_trust_region_history_2d,
    )
    from .plotting_plotly import (
        add_bo_points_plotly,
        plot_function_plotly,
        plot_model_predictions_plotly,
    )
except Exception as e:
    print(
        "trieste.experimental.plotting requires matplotlib and plotly to be installed."
        "\nOne way to do this is to install 'trieste[plotting]'."
    )
    raise e



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains entropy-based acquisition function builders.
"""
from __future__ import annotations

from typing import List, Optional, TypeVar, cast, overload

import tensorflow as tf
import tensorflow_probability as tfp
from typing_extensions import Protocol, runtime_checkable

from ...data import Dataset, add_fidelity_column
from ...models import ProbabilisticModel
from ...models.gpflow.interface import SupportsCovarianceBetweenPoints
from ...models.interfaces import (
    HasTrajectorySampler,
    SupportsCovarianceWithTopFidelity,
    SupportsGetObservationNoise,
    SupportsPredictY,
)
from ...space import SearchSpace
from ...types import TensorType
from ..interface import (
    AcquisitionFunction,
    AcquisitionFunctionClass,
    PenalizationFunction,
    ProbabilisticModelType,
    SingleModelAcquisitionBuilder,
    SingleModelGreedyAcquisitionBuilder,
    UpdatablePenalizationFunction,
)
from ..sampler import ExactThompsonSampler, ThompsonSampler

CLAMP_LB = 1e-8


class MinValueEntropySearch(SingleModelAcquisitionBuilder[ProbabilisticModelType]):
    r"""
    Builder for the max-value entropy search acquisition function modified for objective
    minimisation. :class:`MinValueEntropySearch` estimates the information in the distribution
    of the objective minimum that would be gained by evaluating the objective at a given point.

    This implementation largely follows :cite:`wang2017max` and samples the objective's minimum
    :math:`y^*` across a large set of sampled locations via either a Gumbel sampler, an exact
    Thompson sampler or an approximate random Fourier feature-based Thompson sampler, with the
    Gumbel sampler being the cheapest but least accurate. Default behavior is to use the
    exact Thompson sampler.
    """

    @overload
    def __init__(
        self: "MinValueEntropySearch[ProbabilisticModel]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: None = None,
    ): ...

    @overload
    def __init__(
        self: "MinValueEntropySearch[ProbabilisticModelType]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[ProbabilisticModelType]] = None,
    ): ...

    def __init__(
        self,
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[ProbabilisticModelType]] = None,
    ):
        """
        :param search_space: The global search space over which the optimisation is defined.
        :param num_samples: Number of samples to draw from the distribution over the minimum of the
            objective function.
        :param grid_size: Size of the grid from which to sample the min-values. We recommend
            scaling this with search space dimension.
        :param min_value_sampler: Sampler which samples minimum values.
        :raise tf.errors.InvalidArgumentError: If

            - ``num_samples`` or ``grid_size`` are negative.
        """
        tf.debugging.assert_positive(num_samples)
        tf.debugging.assert_positive(grid_size)

        if min_value_sampler is not None:
            if not min_value_sampler.sample_min_value:
                raise ValueError(
                    """
                    Minvalue Entropy Search requires a min_value_sampler that samples minimum
                    values, however the passed sampler has sample_min_value=False.
                    """
                )
        else:
            min_value_sampler = ExactThompsonSampler(sample_min_value=True)

        self._min_value_sampler = min_value_sampler
        self._search_space = search_space
        self._num_samples = num_samples
        self._grid_size = grid_size

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer.
        :return: The max-value entropy search acquisition function modified for objective
            minimisation. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        query_points = self._search_space.sample(num_samples=self._grid_size)
        tf.debugging.assert_same_float_dtype([dataset.query_points, query_points])
        query_points = tf.concat([dataset.query_points, query_points], 0)
        min_value_samples = self._min_value_sampler.sample(model, self._num_samples, query_points)

        return min_value_entropy_search(model, min_value_samples)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(isinstance(function, min_value_entropy_search), [tf.constant([])])

        query_points = self._search_space.sample(num_samples=self._grid_size)
        tf.debugging.assert_same_float_dtype([dataset.query_points, query_points])
        query_points = tf.concat([dataset.query_points, query_points], 0)
        min_value_samples = self._min_value_sampler.sample(model, self._num_samples, query_points)
        function.update(min_value_samples)  # type: ignore
        return function


class min_value_entropy_search(AcquisitionFunctionClass):
    def __init__(self, model: ProbabilisticModel, samples: TensorType):
        r"""
        Return the max-value entropy search acquisition function (adapted from :cite:`wang2017max`),
        modified for objective minimisation. This function calculates the information gain (or
        change in entropy) in the distribution over the objective minimum :math:`y^*`, if we were
        to evaluate the objective at a given point.

        :param model: The model of the objective function.
        :param samples: Samples from the distribution over :math:`y^*`.
        :return: The max-value entropy search acquisition function modified for objective
            minimisation. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise ValueError or tf.errors.InvalidArgumentError: If ``samples`` has rank less than two,
            or is empty.
        """
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))

        self._model = model
        self._samples = tf.Variable(samples)

    def update(self, samples: TensorType) -> None:
        """Update the acquisition function with new samples."""
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))
        self._samples.assign(samples)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        fmean, fvar = self._model.predict(tf.squeeze(x, -2))
        fsd = tf.math.sqrt(fvar)
        fsd = tf.clip_by_value(
            fsd, CLAMP_LB, fmean.dtype.max
        )  # clip below to improve numerical stability

        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
        gamma = (tf.squeeze(self._samples) - fmean) / fsd

        log_minus_cdf = normal.log_cdf(-gamma)
        ratio = tf.math.exp(normal.log_prob(gamma) - log_minus_cdf)
        f_acqu_x = -gamma * ratio / 2 - log_minus_cdf

        return tf.math.reduce_mean(f_acqu_x, axis=1, keepdims=True)


@runtime_checkable
class SupportsCovarianceObservationNoise(
    SupportsCovarianceBetweenPoints, SupportsGetObservationNoise, Protocol
):
    """A model that supports both covariance_between_points and get_observation_noise."""


class SupportsCovarianceObservationNoiseTrajectory(
    HasTrajectorySampler, SupportsCovarianceObservationNoise, Protocol
):
    """A model that supports covariance_between_points and get_observation_noise, and also
    has an associated trajectory sampler."""


GIBBONModelType = TypeVar(
    "GIBBONModelType", bound=SupportsCovarianceObservationNoise, contravariant=True
)
""" Type variable bound to :class:`~trieste.models.SupportsCovarianceObservationNoise`. """


class GIBBON(SingleModelGreedyAcquisitionBuilder[GIBBONModelType]):
    r"""
    The General-purpose Information-Based Bayesian Optimisation (GIBBON) acquisition function
    of :cite:`Moss:2021`. :class:`GIBBON` provides a computationally cheap approximation of the
    information gained about (i.e the change in entropy of) the objective function's minimum by
    evaluating a batch of candidate points. Batches are built in a greedy manner.

    This implementation follows :cite:`Moss:2021` but is modified for function
    minimisation (rather than maximisation). We sample the objective's minimum
    :math:`y^*` across a large set of sampled locations via either a Gumbel sampler, an exact
    Thompson sampler or an approximate random Fourier feature-based Thompson sampler, with the
    Gumbel sampler being the cheapest but least accurate. Default behavior is to use the
    exact Thompson sampler.
    """

    @overload
    def __init__(
        self: "GIBBON[SupportsCovarianceObservationNoise]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: None = None,
        rescaled_repulsion: bool = True,
    ): ...

    @overload
    def __init__(
        self: "GIBBON[GIBBONModelType]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[GIBBONModelType]] = None,
        rescaled_repulsion: bool = True,
    ): ...

    def __init__(
        self,
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[GIBBONModelType]] = None,
        rescaled_repulsion: bool = True,
    ):
        """
        :param search_space: The global search space over which the optimisation is defined.
        :param num_samples: Number of samples to draw from the distribution over the minimum of
            the objective function.
        :param grid_size: Size of the grid from which to sample the min-values. We recommend
            scaling this with search space dimension.
        :param min_value_sampler: Sampler which samples minimum values.
        :param rescaled_repulsion: If True, then downweight GIBBON's repulsion term to improve
            batch optimization performance.
        :raise tf.errors.InvalidArgumentError: If

            - ``num_samples`` is not positive, or
            - ``grid_size`` is not positive.
        """
        tf.debugging.assert_positive(num_samples)
        tf.debugging.assert_positive(grid_size)

        if min_value_sampler is not None:
            if not min_value_sampler.sample_min_value:
                raise ValueError(
                    """
                    GIBBON requires a min_value_sampler that samples minimum values,
                    however the passed sampler has sample_min_value=False.
                    """
                )
        else:
            min_value_sampler = ExactThompsonSampler(sample_min_value=True)

        self._min_value_sampler = min_value_sampler
        self._search_space = search_space
        self._num_samples = num_samples
        self._grid_size = grid_size
        self._rescaled_repulsion = rescaled_repulsion

        self._min_value_samples: Optional[TensorType] = None
        self._quality_term: Optional[gibbon_quality_term] = None
        self._diversity_term: Optional[gibbon_repulsion_term] = None
        self._gibbon_acquisition: Optional[AcquisitionFunction] = None

    def prepare_acquisition_function(
        self,
        model: GIBBONModelType,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :param pending_points: The points we penalize with respect to.
        :return: The GIBBON acquisition function modified for objective minimisation.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        if not isinstance(model, SupportsCovarianceObservationNoise):
            raise NotImplementedError(
                f"GIBBON only works with models that support "
                f"covariance_between_points and get_observation_noise; received {model!r}"
            )

        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        acq = self._update_quality_term(dataset, model)
        if pending_points is not None and len(pending_points) != 0:
            acq = self._update_repulsion_term(acq, dataset, model, pending_points)

        return acq

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: GIBBONModelType,
        dataset: Optional[Dataset] = None,
        pending_points: Optional[TensorType] = None,
        new_optimization_step: bool = True,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer. Must be populated.
        :param pending_points: Points already chosen to be in the current batch (of shape [M,D]),
            where M is the number of pending points and D is the search space dimension.
        :param new_optimization_step: Indicates whether this call to update_acquisition_function
            is to start of a new optimization step, or to continue collecting batch of points
            for the current step. Defaults to ``True``.
        :return: The updated acquisition function.
        """
        tf.debugging.Assert(dataset is not None, [tf.constant([])])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        tf.debugging.Assert(self._quality_term is not None, [tf.constant([])])

        if new_optimization_step:
            self._update_quality_term(dataset, model)

        if pending_points is None:
            # no repulsion term required if no pending_points.
            return cast(AcquisitionFunction, self._quality_term)

        return self._update_repulsion_term(function, dataset, model, pending_points)

    def _update_repulsion_term(
        self,
        function: Optional[AcquisitionFunction],
        dataset: Dataset,
        model: GIBBONModelType,
        pending_points: Optional[TensorType] = None,
    ) -> AcquisitionFunction:
        tf.debugging.assert_rank(pending_points, 2)

        if self._gibbon_acquisition is not None and isinstance(
            self._diversity_term, gibbon_repulsion_term
        ):
            # if possible, just update the repulsion term
            self._diversity_term.update(pending_points)
            return self._gibbon_acquisition
        else:
            # otherwise construct a new repulsion term and acquisition function
            self._diversity_term = gibbon_repulsion_term(
                model, pending_points, rescaled_repulsion=self._rescaled_repulsion
            )
            self._gibbon_acquisition = GibbonAcquisition(
                cast(AcquisitionFunction, self._quality_term), self._diversity_term
            )
            return self._gibbon_acquisition

    def _update_quality_term(self, dataset: Dataset, model: GIBBONModelType) -> AcquisitionFunction:
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")

        query_points = self._search_space.sample(num_samples=self._grid_size)
        tf.debugging.assert_same_float_dtype([dataset.query_points, query_points])
        query_points = tf.concat([dataset.query_points, query_points], 0)
        self._min_value_samples = self._min_value_sampler.sample(
            model, self._num_samples, query_points
        )

        if self._quality_term is not None:  # if possible, just update the quality term
            self._quality_term.update(self._min_value_samples)
        else:  # otherwise build quality term
            self._quality_term = gibbon_quality_term(model, self._min_value_samples)
        return cast(AcquisitionFunction, self._quality_term)


class GibbonAcquisition:
    """Class representing a GIBBON acquisition function."""

    # (note that this needs to be defined as a top level class make it pickleable)
    def __init__(self, quality_term: AcquisitionFunction, diversity_term: PenalizationFunction):
        """
        :param quality_term: Quality term.
        :param diversity_term: Diversity term.
        """
        self._quality_term = quality_term
        self._diversity_term = diversity_term

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        return self._diversity_term(x) + self._quality_term(x)


class gibbon_quality_term(AcquisitionFunctionClass):
    def __init__(self, model: SupportsCovarianceObservationNoise, samples: TensorType):
        """
        GIBBON's quality term measures the amount of information that each individual
        batch element provides about the objective function's minimal value :math:`y^*` (ensuring
        that evaluations are targeted in promising areas of the space).

        :param model: The model of the objective function. GIBBON requires a model with
            a :method:covariance_between_points method and so GIBBON only
            supports :class:`GaussianProcessRegression` models.
        :param samples: Samples from the distribution over :math:`y^*`.
        :return: GIBBON's quality term. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise ValueError or tf.errors.InvalidArgumentError: If ``samples`` does not have rank two,
            or is empty, or if ``model`` has no homoscedastic observation noise.
        :raise AttributeError: If ``model`` doesn't implement covariance_between_points method.
        """
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))

        try:
            model.get_observation_noise()
        except NotImplementedError:
            raise ValueError(
                """
                GIBBON only currently supports homoscedastic gpflow models
                with a likelihood.variance attribute.
                """
            )

        self._model = model
        self._samples = tf.Variable(samples)

    def update(self, samples: TensorType) -> None:
        """Update the acquisition function with new samples."""
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))
        self._samples.assign(samples)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:  # [N, D] -> [N, 1]
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )

        fmean, fvar = self._model.predict(tf.squeeze(x, -2))
        noise_variance = self._model.get_observation_noise()
        yvar = fvar + tf.cast(noise_variance, fmean.dtype)  # predictive variance of observations

        rho_squared = fvar / yvar  # squared correlation between observations and latent function
        fsd = tf.clip_by_value(
            tf.math.sqrt(fvar), CLAMP_LB, fmean.dtype.max
        )  # clip below to improve numerical stability
        gamma = (tf.squeeze(self._samples) - fmean) / fsd

        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
        log_minus_cdf = normal.log_cdf(-gamma)
        ratio = tf.math.exp(normal.log_prob(gamma) - log_minus_cdf)
        inner_log = 1 + rho_squared * ratio * (gamma - ratio)

        return -0.5 * tf.math.reduce_mean(tf.math.log(inner_log), axis=1, keepdims=True)  # [N, 1]


class gibbon_repulsion_term(UpdatablePenalizationFunction):
    def __init__(
        self,
        model: SupportsCovarianceObservationNoise,
        pending_points: TensorType,
        rescaled_repulsion: bool = True,
    ):
        r"""
        GIBBON's repulsion term encourages diversity within the batch
        (achieving high values for points with low predictive correlation).

        The repulsion term :math:`r=\log |C|` is given by the log determinant of the predictive
        correlation matrix :math:`C` between the `m` pending points and the current candidate.
        The predictive covariance :math:`V` can be expressed as :math:V = [[v, A], [A, B]]` for a
        tensor :math:`B` with shape [`m`,`m`] and so we can efficiently calculate :math:`|V|` using
        the formula for the determinant of block matrices, i.e
        :math:`|V| = (v - A^T * B^{-1} * A) * |B|`.
        Note that when using GIBBON for purely sequential optimization, the repulsion term is
        not required.

        As GIBBON's batches are built in a greedy manner, i.e sequentially adding points to build a
        set of `m` pending points, we need only ever calculate the entropy reduction provided by
        adding the current candidate point to the current pending points, not the full information
        gain provided by evaluating all the pending points. This allows for a modest computational
        saving.

        When performing batch BO, GIBBON's approximation can sometimes become
        less accurate as its repulsion term dominates. Therefore, we follow the
        arguments of :cite:`Moss:2021` and divide GIBBON's repulsion term by :math:`B^{2}`. This
        behavior can be deactivated by setting `rescaled_repulsion` to False.

        :param model: The model of the objective function. GIBBON requires a model with
            a :method:covariance_between_points method and so GIBBON only
            supports :class:`GaussianProcessRegression` models.
        :param pending_points: The points already chosen in the current batch.
        :param rescaled_repulsion: If True, then downweight GIBBON's repulsion term to improve
            batch optimization performance.
        :return: GIBBON's repulsion term. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise ValueError or tf.errors.InvalidArgumentError: If ``pending_points`` does not have
            rank two, or is empty, or if ``model`` has no homoscedastic observation noise.
        :raise AttributeError: If ``model`` doesn't implement covariance_between_points method.
        """
        tf.debugging.assert_rank(pending_points, 2)
        tf.debugging.assert_positive(len(pending_points))

        try:
            model.get_observation_noise()
        except NotImplementedError:
            raise ValueError(
                """
                GIBBON only currently supports homoscedastic gpflow models
                with a likelihood.variance attribute.
                """
            )

        if not hasattr(model, "covariance_between_points"):
            raise AttributeError(
                """
                GIBBON only supports models with a covariance_between_points method.
                """
            )

        self._model = model
        self._pending_points = tf.Variable(pending_points, shape=[None, *pending_points.shape[1:]])
        self._rescaled_repulsion = rescaled_repulsion

    def update(
        self,
        pending_points: TensorType,
        lipschitz_constant: TensorType = None,
        eta: TensorType = None,
    ) -> None:
        """Update the repulsion term with new variable values."""
        self._pending_points.assign(pending_points)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This penalization function cannot be calculated for batches of points.",
        )

        fmean, fvar = self._model.predict(tf.squeeze(x, -2))
        noise_variance = self._model.get_observation_noise()
        yvar = fvar + noise_variance  # need predictive variance of observations

        _, B = self._model.predict_joint(self._pending_points)  # [1, m, m]

        B_shape = tf.shape(B)
        noise = noise_variance * tf.eye(
            B_shape[-2], batch_shape=B_shape[:-2], dtype=B.dtype
        )  # need predictive variance of observations

        L = tf.linalg.cholesky(B + noise)

        A = tf.squeeze(
            tf.expand_dims(
                self._model.covariance_between_points(tf.squeeze(x, -2), self._pending_points),
                axis=-1,
            ),
            axis=0,
        )  # [N, m, 1]
        L_inv_A = tf.linalg.triangular_solve(L, A)
        V_det = yvar - tf.squeeze(
            tf.matmul(L_inv_A, L_inv_A, transpose_a=True), -1
        )  # equation for determinant of block matrices
        repulsion = 0.5 * (tf.math.log(V_det) - tf.math.log(yvar))

        if self._rescaled_repulsion:
            batch_size = tf.cast(tf.shape(self._pending_points)[0], dtype=fmean.dtype)
            repulsion_weight = (1 / batch_size) ** (2)
        else:
            repulsion_weight = 1.0

        return repulsion_weight * repulsion


@runtime_checkable
class SupportsCovarianceWithTopFidelityPredictY(
    SupportsCovarianceWithTopFidelity, SupportsPredictY, Protocol
):
    """A model that is both multifidelity and supports predict_y."""


MUMBOModelType = TypeVar(
    "MUMBOModelType", bound=SupportsCovarianceWithTopFidelityPredictY, contravariant=True
)
""" Type variable bound to :class:`~trieste.models.SupportsCovarianceWithTopFidelityPredictY`. """


class MUMBO(MinValueEntropySearch[MUMBOModelType]):
    r"""
    Builder for the MUlti-task Max-value Bayesian Optimization MUMBO acquisition function modified
    for objective minimisation. :class:`MinValueEntropySearch` estimates the information in the
    distribution of the objective minimum that would be gained by evaluating the objective at a
    given point on a given fidelity level.

    This implementation largely follows :cite:`moss2021mumbo` and samples the objective's minimum
    :math:`y^*` across a large set of sampled locations via either a Gumbel sampler, an exact
    Thompson sampler or an approximate random Fourier feature-based Thompson sampler, with the
    Gumbel sampler being the cheapest but least accurate. Default behavior is to use the
    exact Thompson sampler.
    """

    @overload
    def __init__(
        self: "MUMBO[SupportsCovarianceWithTopFidelityPredictY]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: None = None,
    ): ...

    @overload
    def __init__(
        self: "MUMBO[MUMBOModelType]",
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[MUMBOModelType]] = None,
    ): ...

    def __init__(
        self,
        search_space: SearchSpace,
        num_samples: int = 5,
        grid_size: int = 1000,
        min_value_sampler: Optional[ThompsonSampler[MUMBOModelType]] = None,
    ):
        super().__init__(search_space, num_samples, grid_size, min_value_sampler)

    def prepare_acquisition_function(
        self,
        model: MUMBOModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The multifidelity model.
        :param dataset: The data from the observer.
        :return: The max-value entropy search acquisition function modified for objective
            minimisation. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        min_value_samples = self.get_min_value_samples_on_top_fidelity(model, dataset)
        return mumbo(model, min_value_samples)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: MUMBOModelType,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer.
        """
        tf.debugging.Assert(dataset is not None, [])
        dataset = cast(Dataset, dataset)
        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
        min_value_samples = self.get_min_value_samples_on_top_fidelity(model, dataset)
        function.update(min_value_samples)  # type: ignore
        return function

    def get_min_value_samples_on_top_fidelity(
        self, model: MUMBOModelType, dataset: Dataset
    ) -> TensorType:
        """
        :param model: The model.
        :param dataset: The data from the observer.
        """
        query_points = self._search_space.sample(num_samples=self._grid_size)
        tf.debugging.assert_same_float_dtype([dataset.query_points, query_points])
        query_points = tf.concat([dataset.query_points, query_points], 0)
        query_points_on_top_fidelity = add_fidelity_column(
            query_points[:, :-1], model.num_fidelities - 1
        )
        return self._min_value_sampler.sample(
            model, self._num_samples, query_points_on_top_fidelity
        )


class mumbo(AcquisitionFunctionClass):
    def __init__(self, model: MUMBOModelType, samples: TensorType):
        r"""
        The MUMBO acquisition function of :cite:`moss2021mumbo`, modified for objective
        minimisation. This function calculates the information gain (or change in entropy) in the
        distribution over the objective minimum :math:`y^*`, if we were to evaluate the objective
        at a given point on a given fidelity level.

        To speed up calculations, we use a trick from :cite:`Moss:2021` and use moment-matching to
        calculate MUMBO's entropy terms rather than numerical integration.

        :param model: The model of the objective function.
        :param samples: Samples from the distribution over :math:`y^*`.
        :return: The MUMBO acquisition function modified for objective
            minimisation. This function will raise :exc:`ValueError` or
            :exc:`~tf.errors.InvalidArgumentError` if used with a batch size greater than one.
        :raise ValueError or tf.errors.InvalidArgumentError: If ``samples`` has rank less than two,
            or is empty.
        """
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))

        self._model = model
        self._samples = tf.Variable(samples)

    def update(self, samples: TensorType) -> None:
        """Update the acquisition function with new samples."""
        tf.debugging.assert_rank(samples, 2)
        tf.debugging.assert_positive(len(samples))
        self._samples.assign(samples)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )

        x_squeezed = tf.squeeze(x, -2)
        x_on_top_fidelity = add_fidelity_column(x_squeezed[:, :-1], self._model.num_fidelities - 1)

        fmean, fvar = self._model.predict(x_on_top_fidelity)
        fsd = tf.clip_by_value(
            tf.math.sqrt(fvar), CLAMP_LB, fmean.dtype.max
        )  # clip below to improve numerical stability
        ymean, yvar = self._model.predict_y(x_squeezed)
        cov = self._model.covariance_with_top_fidelity(x_squeezed)

        # calculate squared correlation between observations and high-fidelity latent function
        rho_squared = (cov**2) / (fvar * yvar)
        rho_squared = tf.clip_by_value(rho_squared, 0.0, 1.0)

        normal = tfp.distributions.Normal(tf.constant(0, fmean.dtype), tf.constant(1, fmean.dtype))
        gamma = (tf.squeeze(self._samples) - fmean) / fsd
        log_minus_cdf = normal.log_cdf(-gamma)
        ratio = tf.math.exp(normal.log_prob(gamma) - log_minus_cdf)

        inner_log = 1 + rho_squared * ratio * (gamma - ratio)

        return -0.5 * tf.math.reduce_mean(tf.math.log(inner_log), axis=1, keepdims=True)  # [N, 1]


class CostWeighting(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    def __init__(self, fidelity_costs: List[float]):
        """
        Builder for a cost-weighted acquisition function which returns the reciprocal of the cost
        associated with the fidelity of each input.

        Note that the fidelity level is assumed to be contained in the inputs final dimension.

        The primary use of this acquisition function is to be used as a product with
        multi-fidelity acquisition functions.
        """

        self._fidelity_costs = fidelity_costs
        self._num_fidelities = len(self._fidelity_costs)

    def prepare_acquisition_function(
        self, model: ProbabilisticModel, dataset: Optional[Dataset] = None
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: The data from the observer. Not actually used here.
        :return: The reciprocal of the costs corresponding to the fidelity level of each input.
        """

        @tf.function
        def acquisition(x: TensorType) -> TensorType:
            tf.debugging.assert_shapes(
                [(x, [..., 1, None])],
                message="This acquisition function only supports batch sizes of one.",
            )
            fidelities = x[..., -1]  # [..., 1]
            tf.debugging.assert_greater(
                tf.cast(self._num_fidelities, fidelities.dtype),
                tf.reduce_max(fidelities),
                message="You are trying to use more fidelity levels than cost levels.",
            )

            costs = tf.gather(self._fidelity_costs, tf.cast(fidelities, tf.int32))

            return tf.cast(1.0 / costs, x.dtype)  # [N, 1]

        return acquisition

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        Nothing to do here, so just return previous cost function.

        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: The data from the observer.
        """
        return function



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions convenient for creating :class:`Observer` objects that return data
from objective functions, appropriately formatted for usage with the toolbox.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Mapping, Optional, Union, overload

from check_shapes import check_shapes

from ..data import Dataset
from ..observer import OBJECTIVE, MultiObserver, Observer, SingleObserver
from ..types import Tag, TensorType
from ..utils.misc import LocalizedTag, flatten_leading_dims


@overload
def mk_observer(objective: Callable[[TensorType], TensorType]) -> SingleObserver: ...


@overload
def mk_observer(objective: Callable[[TensorType], TensorType], key: Tag) -> MultiObserver: ...


def mk_observer(
    objective: Callable[[TensorType], TensorType], key: Optional[Tag] = None
) -> Observer:
    """
    :param objective: An objective function designed to be used with a single data set and model.
    :param key: An optional key to use to access the data from the observer result.
    :return: An observer returning the data from ``objective``.
    """
    if key is not None:
        return lambda qp: {key: Dataset(qp, objective(qp))}
    else:
        return lambda qp: Dataset(qp, objective(qp))


def mk_multi_observer(**kwargs: Callable[[TensorType], TensorType]) -> MultiObserver:
    """
    :param kwargs: Observation functions.
    :return: An multi-observer returning the data from ``kwargs``.
    """
    return lambda qp: {key: Dataset(qp, objective(qp)) for key, objective in kwargs.items()}


def mk_batch_observer(
    objective_or_observer: Union[Callable[[TensorType], TensorType], Observer],
    default_key: Tag = OBJECTIVE,
) -> MultiObserver:
    """
    Create an observer that returns the data from ``objective`` or an existing ``observer``
    separately for each query point in a batch.

    :param objective_or_observer: An objective or an existing observer.
    :param default_key: The default key to use if ``objective_or_observer`` is an objective or
        does not return a mapping.
    :return: A multi-observer across the batch dimension of query points, returning the data from
        ``objective_or_observer``.
    """

    @check_shapes("qps: [n_points, batch_size, n_dims]")
    def _observer(qps: TensorType) -> Mapping[Tag, Dataset]:
        # Call objective with rank 2 query points by flattening batch dimension.
        # Some objectives might only expect rank 2 query points, so this is safer.
        batch_size = qps.shape[1]
        flat_qps, unflatten = flatten_leading_dims(qps)
        obs_or_dataset = objective_or_observer(flat_qps)

        if not isinstance(obs_or_dataset, (Mapping, Dataset)):
            # Just a single observation, so wrap in a dataset.
            obs_or_dataset = Dataset(flat_qps, obs_or_dataset)

        if isinstance(obs_or_dataset, Dataset):
            # Convert to a mapping with a default key.
            obs_or_dataset = {default_key: obs_or_dataset}

        datasets = {}
        for key, dataset in obs_or_dataset.items():
            # Include overall dataset and per batch dataset.
            flat_obs = dataset.observations
            qps = unflatten(flat_qps)
            obs = unflatten(flat_obs)
            datasets[key] = dataset
            for i in range(batch_size):
                datasets[LocalizedTag(key, i)] = Dataset(qps[:, i], obs[:, i])

        return datasets

    return _observer



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the Ask/Tell API for users of Trieste who would like to
perform Bayesian Optimization with external control of the optimization loop.
"""


from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, Generic, Mapping, Optional, Sequence, Type, TypeVar, cast, overload

import tensorflow as tf

from .models.utils import optimize_model_and_save_result

try:
    import pandas as pd
except ModuleNotFoundError:
    pd = None

from . import logging
from .acquisition.rule import (
    AcquisitionRule,
    EfficientGlobalOptimization,
    LocalDatasetsAcquisitionRule,
)
from .acquisition.utils import with_local_datasets
from .bayesian_optimizer import (
    FrozenRecord,
    OptimizationResult,
    Record,
    observation_plot_init,
    write_summary_initial_model_fit,
    write_summary_observations,
    write_summary_query_points,
)
from .data import Dataset
from .models import ProbabilisticModel, TrainableProbabilisticModel
from .observer import OBJECTIVE
from .space import SearchSpace
from .types import State, Tag, TensorType
from .utils import Ok, Timer
from .utils.misc import LocalizedTag, get_value_for_tag, ignoring_local_tags

StateType = TypeVar("StateType")
""" Unbound type variable. """

SearchSpaceType = TypeVar("SearchSpaceType", bound=SearchSpace)
""" Type variable bound to :class:`SearchSpace`. """

ProbabilisticModelType = TypeVar(
    "ProbabilisticModelType", bound=ProbabilisticModel, contravariant=True
)
""" Contravariant type variable bound to :class:`ProbabilisticModel`. """

AskTellOptimizerType = TypeVar("AskTellOptimizerType")


@dataclass(frozen=True)
class AskTellOptimizerState(Generic[StateType, ProbabilisticModelType]):
    """
    Internal state for an Ask/Tell optimizer. This can be obtained using the optimizer's
    `to_state` method, and can be used to initialise a new instance of the optimizer.
    """

    record: Record[StateType, ProbabilisticModelType]
    """ A record of the current state of the optimization. """

    local_data_ixs: Optional[Sequence[TensorType]]
    """ Indices to the local data, for LocalDatasetsAcquisitionRule rules
    when `track_data` is `False`. """

    local_data_len: Optional[int]
    """ Length of the datasets, for LocalDatasetsAcquisitionRule rules
    when `track_data` is `False`. """


class AskTellOptimizerABC(ABC, Generic[SearchSpaceType, ProbabilisticModelType]):
    """
    This class provides Ask/Tell optimization interface. It is designed for those use cases
    when control of the optimization loop by Trieste is impossible or not desirable.
    For the default use case with model training, refer to :class:`AskTellOptimizer`.
    For more details about the Bayesian Optimization routine, refer to :class:`BayesianOptimizer`.
    """

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, ProbabilisticModelType],
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, ProbabilisticModelType],
        acquisition_rule: AcquisitionRule[TensorType, SearchSpaceType, ProbabilisticModelType],
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Mapping[Tag, Dataset],
        models: Mapping[Tag, ProbabilisticModelType],
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, ProbabilisticModelType
        ],
        acquisition_state: StateType | None,
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Dataset,
        models: ProbabilisticModelType,
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Dataset,
        models: ProbabilisticModelType,
        acquisition_rule: AcquisitionRule[TensorType, SearchSpaceType, ProbabilisticModelType],
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    @overload
    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Dataset,
        models: ProbabilisticModelType,
        acquisition_rule: AcquisitionRule[
            State[StateType | None, TensorType], SearchSpaceType, ProbabilisticModelType
        ],
        acquisition_state: StateType | None = None,
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ): ...

    def __init__(
        self,
        search_space: SearchSpaceType,
        datasets: Mapping[Tag, Dataset] | Dataset,
        models: Mapping[Tag, ProbabilisticModelType] | ProbabilisticModelType,
        acquisition_rule: (
            AcquisitionRule[
                TensorType | State[StateType | None, TensorType],
                SearchSpaceType,
                ProbabilisticModelType,
            ]
            | None
        ) = None,
        acquisition_state: StateType | None = None,
        *,
        fit_model: bool = True,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ):
        """
        :param search_space: The space over which to search for the next query point.
        :param datasets: Already observed input-output pairs for each tag.
        :param models: The model to use for each :class:`~trieste.data.Dataset` in
            ``datasets``.
        :param acquisition_rule: The acquisition rule, which defines how to search for a new point
            on each optimization step. Defaults to
            :class:`~trieste.acquisition.rule.EfficientGlobalOptimization` with default
            arguments. Note that if the default is used, this implies the tags must be
            `OBJECTIVE` and the search space can be any :class:`~trieste.space.SearchSpace`.
        :param acquisition_state: The optional acquisition state for stateful acquisitions.
        :param fit_model: If `True` (default), models passed in will be optimized on the given data.
            If `False`, the models are assumed to be optimized already.
        :param track_data: If `True` (default), the optimizer will track the changing
            datasets via a local copy. If `False`, it will infer new datasets from
            updates to the global datasets (optionally using `local_data_ixs` and indices passed
            in to `tell`).
        :param local_data_ixs: Indices to the local data in the initial datasets. If unspecified,
            assumes that the initial datasets are global.
        :param local_data_len: Optional length of the data when the passed in `local_data_ixs`
            were measured. If the data has increased since then, the indices are extended.
        :raise ValueError: If any of the following are true:
            - the keys in ``datasets`` and ``models`` do not match
            - ``datasets`` or ``models`` are empty
            - default acquisition is used but incompatible with other inputs
        """
        self._search_space = search_space
        self._acquisition_record = self._acquisition_state = acquisition_state

        if not datasets or not models:
            raise ValueError("dicts of datasets and models must be populated.")

        if isinstance(datasets, Dataset):
            datasets = {OBJECTIVE: datasets}
        if not isinstance(models, Mapping):
            models = {OBJECTIVE: models}

        # reassure the type checker that everything is tagged
        datasets = cast(Dict[Tag, Dataset], datasets)
        models = cast(Dict[Tag, ProbabilisticModelType], models)

        # Get set of dataset and model keys, ignoring any local tag index. That is, only the
        # global tag part is considered.
        datasets_keys = {LocalizedTag.from_tag(tag).global_tag for tag in datasets.keys()}
        models_keys = {LocalizedTag.from_tag(tag).global_tag for tag in models.keys()}
        if datasets_keys != models_keys:
            raise ValueError(
                f"datasets and models should contain the same keys. Got {datasets_keys} and"
                f" {models_keys} respectively."
            )

        self._datasets = datasets
        self._models = models
        self.track_data = track_data

        self._query_plot_dfs: dict[int, pd.DataFrame] = {}
        self._observation_plot_dfs = observation_plot_init(self._datasets)

        if acquisition_rule is None:
            if self._datasets.keys() != {OBJECTIVE}:
                raise ValueError(
                    f"Default acquisition rule EfficientGlobalOptimization requires tag"
                    f" {OBJECTIVE!r}, got keys {self._datasets.keys()}"
                )

            self._acquisition_rule = cast(
                AcquisitionRule[TensorType, SearchSpaceType, ProbabilisticModelType],
                EfficientGlobalOptimization(),
            )
        else:
            self._acquisition_rule = acquisition_rule

        if isinstance(self._acquisition_rule, LocalDatasetsAcquisitionRule):
            # In order to support local datasets, account for the case where there may be an initial
            # dataset that is not tagged per region. In this case, only the global dataset will
            # exist in datasets. We want to copy this initial dataset to all the regions.
            num_local_datasets = self._acquisition_rule.num_local_datasets
            if self.track_data:
                datasets = self._datasets = with_local_datasets(self._datasets, num_local_datasets)
            else:
                self._dataset_len = self.dataset_len(self._datasets)
                if local_data_ixs is not None:
                    self._dataset_ixs = list(local_data_ixs)
                    if local_data_len is not None:
                        # infer new dataset indices from change in dataset sizes
                        num_new_points = self._dataset_len - local_data_len
                        if num_new_points < 0 or (
                            num_local_datasets > 0 and num_new_points % num_local_datasets != 0
                        ):
                            raise ValueError(
                                "Cannot infer new data points as datasets haven't increased by "
                                f"a multiple of {num_local_datasets}"
                            )
                        for i in range(num_local_datasets):
                            self._dataset_ixs[i] = tf.concat(
                                [
                                    self._dataset_ixs[i],
                                    tf.range(0, num_new_points, num_local_datasets)
                                    + local_data_len
                                    + i,
                                ],
                                -1,
                            )
                else:
                    self._dataset_ixs = [
                        tf.range(self._dataset_len) for _ in range(num_local_datasets)
                    ]

                datasets = with_local_datasets(
                    self._datasets, num_local_datasets, self._dataset_ixs
                )
            self._acquisition_rule.initialize_subspaces(search_space)

        filtered_datasets: (
            Mapping[Tag, Dataset] | State[StateType | None, Mapping[Tag, Dataset]]
        ) = self._acquisition_rule.filter_datasets(self._models, datasets)
        if callable(filtered_datasets):
            self._acquisition_state, self._filtered_datasets = filtered_datasets(
                self._acquisition_state
            )
        else:
            self._filtered_datasets = filtered_datasets

        if fit_model:
            with Timer() as initial_model_fitting_timer:
                for tag, model in self._models.items():
                    # Prefer local dataset if available.
                    tags = [tag, LocalizedTag.from_tag(tag).global_tag]
                    _, dataset = get_value_for_tag(self._filtered_datasets, *tags)
                    assert dataset is not None
                    self.update_model(model, dataset)

            summary_writer = logging.get_tensorboard_writer()
            if summary_writer:
                with summary_writer.as_default(step=logging.get_step_number()):
                    write_summary_initial_model_fit(
                        self._datasets, self._models, initial_model_fitting_timer
                    )

    @abstractmethod
    def update_model(self, model: ProbabilisticModelType, dataset: Dataset) -> None:
        """
        Update the model on the specified dataset, for example by training.
        Called during the Tell stage and optionally at initial fitting.
        """

    def __repr__(self) -> str:
        """Print-friendly string representation"""
        return f"""AskTellOptimizer({self._search_space!r}, {self._datasets!r},
               {self._models!r}, {self._acquisition_rule!r}), "
               {self._acquisition_state!r}"""

    @property
    def datasets(self) -> Mapping[Tag, Dataset]:
        """The current datasets."""
        return self._datasets

    @property
    def dataset(self) -> Dataset:
        """The current dataset when there is just one dataset."""
        # Ignore local datasets.
        datasets: Mapping[Tag, Dataset] = ignoring_local_tags(self.datasets)
        if len(datasets) == 1:
            return next(iter(datasets.values()))
        else:
            raise ValueError(f"Expected a single dataset, found {len(datasets)}")

    @property
    def local_data_ixs(self) -> Optional[Sequence[TensorType]]:
        """Indices to the local data. Only stored for LocalDatasetsAcquisitionRule rules
        when `track_data` is `False`."""
        if isinstance(self._acquisition_rule, LocalDatasetsAcquisitionRule) and not self.track_data:
            return self._dataset_ixs
        return None

    @property
    def local_data_len(self) -> Optional[int]:
        """Data length. Only stored for LocalDatasetsAcquisitionRule rules
        when `track_data` is `False`."""
        if isinstance(self._acquisition_rule, LocalDatasetsAcquisitionRule) and not self.track_data:
            return self._dataset_len
        return None

    @property
    def models(self) -> Mapping[Tag, ProbabilisticModelType]:
        """The current models."""
        return self._models

    @models.setter
    def models(self, models: Mapping[Tag, ProbabilisticModelType]) -> None:
        """Update the current models."""
        if models.keys() != self.models.keys():
            raise ValueError(
                f"New models contain incorrect keys. Expected {self.models.keys()}, "
                f"received {models.keys()}."
            )
        self._models = dict(models)

    @property
    def model(self) -> ProbabilisticModel:
        """The current model when there is just one model."""
        # Ignore local models.
        models: Mapping[Tag, ProbabilisticModel] = ignoring_local_tags(self.models)
        if len(models) == 1:
            return next(iter(models.values()))
        else:
            raise ValueError(f"Expected a single model, found {len(models)}")

    @model.setter
    def model(self, model: ProbabilisticModelType) -> None:
        """Update the current model, using the OBJECTIVE tag."""
        if len(self.models) != 1:
            raise ValueError(f"Expected a single model, found {len(self.models)}")
        if self.models.keys() != {OBJECTIVE}:
            raise ValueError(
                f"Expected a single model tagged OBJECTIVE, found {self.models.keys()}. "
                "To update this, pass in a dictionary to the models property instead."
            )
        self._models = {OBJECTIVE: model}

    @property
    def acquisition_state(self) -> StateType | None:
        """The current acquisition state."""
        return self._acquisition_state

    @classmethod
    def dataset_len(cls, datasets: Mapping[Tag, Dataset]) -> int:
        """Helper method for inferring the global dataset size."""
        dataset_lens = {
            tag: int(tf.shape(dataset.query_points)[0])
            for tag, dataset in datasets.items()
            if not LocalizedTag.from_tag(tag).is_local
        }
        unique_lens, _ = tf.unique(list(dataset_lens.values()))
        if len(unique_lens) == 1:
            return int(unique_lens[0])
        else:
            raise ValueError(
                f"Expected unique global dataset size, got {unique_lens}: {dataset_lens}"
            )

    @classmethod
    def from_record(
        cls: Type[AskTellOptimizerType],
        record: (
            Record[StateType, ProbabilisticModelType]
            | FrozenRecord[StateType, ProbabilisticModelType]
        ),
        search_space: SearchSpaceType,
        acquisition_rule: (
            AcquisitionRule[
                TensorType | State[StateType | None, TensorType],
                SearchSpaceType,
                ProbabilisticModelType,
            ]
            | None
        ) = None,
        track_data: bool = True,
        local_data_ixs: Optional[Sequence[TensorType]] = None,
        local_data_len: Optional[int] = None,
    ) -> AskTellOptimizerType:
        """Creates new :class:`~AskTellOptimizer` instance from provided optimization state.
        Model training isn't triggered upon creation of the instance.

        :param record: Optimization state record.
        :param search_space: The space over which to search for the next query point.
        :param acquisition_rule: The acquisition rule, which defines how to search for a new point
            on each optimization step. Defaults to
            :class:`~trieste.acquisition.rule.EfficientGlobalOptimization` with default
            arguments.
        :param track_data: Whether the optimizer tracks the changing datasets via a local copy.
        :param local_data_ixs: Indices to local data for local rules with `track_data` False.
        :param local_data_len: Original data length for local rules with `track_data` False.
        :return: New instance of :class:`~AskTellOptimizer`.
        """
        # we are recovering previously saved optimization state
        # so the model was already trained
        # thus there is no need to train it again

        # type ignore below is because this relies on subclasses not overriding __init__
        # ones that do may also need to override this to get it to work
        return cls(  # type: ignore
            search_space,
            record.datasets,
            record.models,
            acquisition_rule=acquisition_rule,
            acquisition_state=record.acquisition_state,
            fit_model=False,
            track_data=track_data,
            local_data_ixs=local_data_ixs,
            local_data_len=local_data_len,
        )

    def to_record(self, copy: bool = True) -> Record[StateType, ProbabilisticModelType]:
        """Collects the current state of the optimization, which includes datasets,
        models and acquisition state (if applicable).

        :param copy: Whether to return a copy of the current state or the original. Copying
            is not supported for all model types. However, continuing the optimization will
            modify the original state.
        :return: An optimization state record.
        """
        try:
            datasets_copy = deepcopy(self._datasets) if copy else self._datasets
            models_copy = deepcopy(self._models) if copy else self._models
            # use the state as it was at acquisition time, not the one modified in
            # filter_datasets in preparation for the next acquisition, so we can reinitialise
            # the AskTellOptimizer using the record
            state_copy = deepcopy(self._acquisition_record) if copy else self._acquisition_record
        except Exception as e:
            raise NotImplementedError(
                "Failed to copy the optimization state. Some models do not support "
                "deecopying (this is particularly common for deep neural network models). "
                "For these models, the `copy` argument of the `to_record` or `to_result` "
                "methods should be set to `False`. This means that the returned state may be "
                "modified by subsequent optimization."
            ) from e

        return Record(datasets=datasets_copy, models=models_copy, acquisition_state=state_copy)

    def to_result(self, copy: bool = True) -> OptimizationResult[StateType, ProbabilisticModelType]:
        """Converts current state of the optimization
        into a :class:`~trieste.data.OptimizationResult` object.

        :param copy: Whether to return a copy of the current state or the original. Copying
            is not supported for all model types. However, continuing the optimization will
            modify the original state.
        :return: A :class:`~trieste.data.OptimizationResult` object.
        """
        record: Record[StateType, ProbabilisticModelType] = self.to_record(copy=copy)
        return OptimizationResult(Ok(record), [])

    @classmethod
    def from_state(
        cls: Type[AskTellOptimizerType],
        state: AskTellOptimizerState[StateType, ProbabilisticModelType],
        search_space: SearchSpaceType,
        acquisition_rule: (
            AcquisitionRule[
                TensorType | State[StateType | None, TensorType],
                SearchSpaceType,
                ProbabilisticModelType,
            ]
            | None
        ) = None,
        track_data: bool = True,
    ) -> AskTellOptimizerType:
        """Creates new :class:`~AskTellOptimizer` instance from provided AskTellOptimizer state.
        Model training isn't triggered upon creation of the instance.

        :param state: AskTellOptimizer state.
        :param search_space: The space over which to search for the next query point.
        :param acquisition_rule: The acquisition rule, which defines how to search for a new point
            on each optimization step. Defaults to
            :class:`~trieste.acquisition.rule.EfficientGlobalOptimization` with default
            arguments.
        :param track_data: Whether the optimizer tracks the changing datasets via a local copy.
        :return: New instance of :class:`~AskTellOptimizer`.
        """
        return cls.from_record(  # type: ignore
            state.record,
            search_space,
            acquisition_rule,
            track_data=track_data,
            local_data_ixs=state.local_data_ixs,
            local_data_len=state.local_data_len,
        )

    def to_state(
        self, copy: bool = False
    ) -> AskTellOptimizerState[StateType, ProbabilisticModelType]:
        """Returns the AskTellOptimizer state, comprising the current optimization state
        alongside any internal AskTellOptimizer state.

        :param copy: Whether to return a copy of the current state or the original. Copying
            is not supported for all model types. However, continuing the optimization will
            modify the original state.
        :return: An :class:`AskTellOptimizerState` object.
        """
        return AskTellOptimizerState(
            record=self.to_record(copy=copy),
            local_data_ixs=self.local_data_ixs,
            local_data_len=self.local_data_len,
        )

    def ask(self) -> TensorType:
        """Suggests a point (or points in batch mode) to observe by optimizing the acquisition
        function. If the acquisition is stateful, its state is saved.

        :return: A :class:`TensorType` instance representing suggested point(s).
        """
        # This trick deserves a comment to explain what's going on
        # acquisition_rule.acquire can return different things:
        # - when acquisition has no state attached, it returns just points
        # - when acquisition has state, it returns a Callable
        #   which, when called, returns state and points
        # so code below is needed to cater for both cases

        with Timer() as query_point_generation_timer:
            points_or_stateful = self._acquisition_rule.acquire(
                self._search_space, self._models, datasets=self._filtered_datasets
            )

        if callable(points_or_stateful):
            self._acquisition_state, query_points = points_or_stateful(self._acquisition_state)
            # also keep a copy of the state to return in to_record
            self._acquisition_record = self._acquisition_state
        else:
            query_points = points_or_stateful

        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                write_summary_query_points(
                    self._datasets,
                    self._models,
                    self._search_space,
                    query_points,
                    query_point_generation_timer,
                    self._query_plot_dfs,
                )

        return query_points

    def tell(
        self,
        new_data: Mapping[Tag, Dataset] | Dataset,
        new_data_ixs: Optional[Sequence[TensorType]] = None,
    ) -> None:
        """Updates optimizer state with new data.

        :param new_data: New observed data. If `track_data` is `False`, this refers to all
            the data.
        :param new_data_ixs: Indices to the new observed local data, if `track_data` is `False`.
            If unspecified, inferred from the change in dataset sizes.
        :raise ValueError: If keys in ``new_data`` do not match those in already built dataset.
        """
        if isinstance(new_data, Dataset):
            new_data = {OBJECTIVE: new_data}

        # The datasets must have the same keys as the existing datasets. Only exception is if
        # the existing datasets are all global and the new data contains local datasets too.
        if all(LocalizedTag.from_tag(tag).local_index is None for tag in self._datasets.keys()):
            global_old = {LocalizedTag.from_tag(tag).global_tag for tag in self._datasets.keys()}
            global_new = {LocalizedTag.from_tag(tag).global_tag for tag in new_data.keys()}
            if global_new != global_old:
                raise ValueError(
                    f"new_data global keys {global_new} doesn't "
                    f"match dataset global keys {global_old}"
                )
        elif self._datasets.keys() != new_data.keys():
            raise ValueError(
                f"new_data keys {new_data.keys()} doesn't "
                f"match dataset keys {self._datasets.keys()}"
            )

        if self.track_data:
            for tag, new_dataset in new_data.items():
                self._datasets[tag] += new_dataset
            datasets: Mapping[Tag, Dataset] = self._datasets
        elif not isinstance(self._acquisition_rule, LocalDatasetsAcquisitionRule):
            datasets = new_data
        else:
            num_local_datasets = len(self._dataset_ixs)
            if new_data_ixs is None:
                # infer dataset indices from change in dataset sizes
                new_dataset_len = self.dataset_len(new_data)
                num_new_points = new_dataset_len - self._dataset_len
                if num_new_points < 0 or (
                    num_local_datasets > 0 and num_new_points % num_local_datasets != 0
                ):
                    raise ValueError(
                        "Cannot infer new data points as datasets haven't increased by "
                        f"a multiple of {num_local_datasets}"
                    )
                for i in range(num_local_datasets):
                    self._dataset_ixs[i] = tf.concat(
                        [
                            self._dataset_ixs[i],
                            tf.range(0, num_new_points, num_local_datasets) + self._dataset_len + i,
                        ],
                        -1,
                    )
            else:
                # use explicit indices
                if len(new_data_ixs) != num_local_datasets:
                    raise ValueError(
                        f"new_data_ixs has {len(new_data_ixs)} entries, "
                        f"expected {num_local_datasets}"
                    )
                for i in range(num_local_datasets):
                    self._dataset_ixs[i] = tf.concat([self._dataset_ixs[i], new_data_ixs[i]], -1)
            datasets = with_local_datasets(new_data, num_local_datasets, self._dataset_ixs)
            self._dataset_len = self.dataset_len(datasets)

        filtered_datasets = self._acquisition_rule.filter_datasets(self._models, datasets)
        if callable(filtered_datasets):
            self._acquisition_state, self._filtered_datasets = filtered_datasets(
                self._acquisition_state
            )
        else:
            self._filtered_datasets = filtered_datasets

        with Timer() as model_fitting_timer:
            for tag, model in self._models.items():
                # Always use the matching dataset to the model. If the model is
                # local, then the dataset should be too by this stage.
                dataset = self._filtered_datasets[tag]
                self.update_model(model, dataset)

        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                write_summary_observations(
                    datasets,
                    self._models,
                    new_data,
                    model_fitting_timer,
                    self._observation_plot_dfs,
                )


TrainableProbabilisticModelType = TypeVar(
    "TrainableProbabilisticModelType", bound=TrainableProbabilisticModel, contravariant=True
)
""" Contravariant type variable bound to :class:`TrainableProbabilisticModel`. """


class AskTellOptimizer(AskTellOptimizerABC[SearchSpaceType, TrainableProbabilisticModelType]):
    """
    This class provides Ask/Tell optimization interface with the default model training
    using the TrainableProbabilisticModel interface.
    """

    def update_model(self, model: TrainableProbabilisticModelType, dataset: Dataset) -> None:
        model.update(dataset)
        optimize_model_and_save_result(model, dataset)


class AskTellOptimizerNoTraining(AskTellOptimizerABC[SearchSpaceType, ProbabilisticModelType]):
    """
    This class provides Ask/Tell optimization interface with no model training performed
    during the Tell stage or at initialization.
    """

    def update_model(self, model: ProbabilisticModelType, dataset: Dataset) -> None:
        pass



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any, Callable, Mapping, Optional

import dill
import gpflow
import tensorflow as tf
from gpflow.inducing_variables import InducingPoints
from gpflow.keras import tf_keras
from gpflux.layers import GPLayer, LatentVariableLayer
from gpflux.models import DeepGP
from tensorflow.python.keras.callbacks import Callback

from ... import logging
from ...data import Dataset
from ...space import EncoderFunction
from ...types import TensorType
from ..interfaces import (
    EncodedTrainableProbabilisticModel,
    HasReparamSampler,
    HasTrajectorySampler,
    ReparametrizationSampler,
    TrajectorySampler,
)
from ..optimizer import KerasOptimizer
from ..utils import (
    write_summary_data_based_metrics,
    write_summary_kernel_parameters,
    write_summary_likelihood_parameters,
)
from .interface import GPfluxPredictor
from .sampler import (
    DeepGaussianProcessDecoupledTrajectorySampler,
    DeepGaussianProcessReparamSampler,
)


class DeepGaussianProcess(
    GPfluxPredictor, EncodedTrainableProbabilisticModel, HasReparamSampler, HasTrajectorySampler
):
    """
    A :class:`TrainableProbabilisticModel` wrapper for a GPflux :class:`~gpflux.models.DeepGP` with
    :class:`GPLayer` or :class:`LatentVariableLayer`: this class does not support e.g. keras layers.
    We provide simple architectures that can be used with this class in the `architectures.py` file.
    """

    def __init__(
        self,
        model: DeepGP | Callable[[], DeepGP],
        optimizer: KerasOptimizer | None = None,
        num_rff_features: int = 1000,
        continuous_optimisation: bool = True,
        compile_args: Optional[Mapping[str, Any]] = None,
        encoder: EncoderFunction | None = None,
    ):
        """
        :param model: The underlying GPflux deep Gaussian process model. Passing in a named closure
            rather than a model can help when copying or serialising.
        :param optimizer: The optimizer wrapper with necessary specifications for compiling and
            training the model. Defaults to :class:`~trieste.models.optimizer.KerasOptimizer` with
            :class:`~tf.optimizers.Adam` optimizer, mean squared error metric and a dictionary of
            default arguments for the Keras `fit` method: 400 epochs, batch size of 1000, and
            verbose 0. A custom callback that reduces the optimizer learning rate is used as well.
            See https://keras.io/api/models/model_training_apis/#fit-method for a list of possible
            arguments.
        :param num_rff_features: The number of random Fourier features used to approximate the
            kernel when calling :meth:`trajectory_sampler`. We use a default of 1000 as it typically
            performs well for a wide range of kernels. Note that very smooth kernels (e.g. RBF) can
            be well-approximated with fewer features.
        :param continuous_optimisation: if True (default), the optimizer will keep track of the
            number of epochs across BO iterations and use this number as initial_epoch. This is
            essential to allow monitoring of model training across BO iterations.
        :param compile_args: Keyword arguments to pass to the ``compile`` method of the
            Keras model (:class:`~tf.keras.Model`).
            See https://keras.io/api/models/model_training_apis/#compile-method for a
            list of possible arguments. The ``optimizer`` and ``metrics`` arguments
            must not be included.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        :raise ValueError: If ``model`` has unsupported layers, ``num_rff_features`` is less than 0,
            if the ``optimizer`` is not of a supported type, or `compile_args` contains
            disallowed arguments.
        """
        if compile_args is None:
            compile_args = {}

        if not {"optimizer", "metrics"}.isdisjoint(compile_args):
            raise ValueError(
                "optimizer and metrics arguments must not be included in compile_args."
            )

        if isinstance(model, DeepGP):
            self._model_closure = None
        else:
            self._model_closure = model
            model = model()

        for layer in model.f_layers:
            if not isinstance(layer, (GPLayer, LatentVariableLayer)):
                raise ValueError(
                    f"`DeepGaussianProcess` can only be built out of `GPLayer` or"
                    f"`LatentVariableLayer`, received {type(layer)} instead."
                )

        super().__init__(optimizer, encoder)

        if num_rff_features <= 0:
            raise ValueError(
                f"num_rff_features must be greater or equal to zero, got {num_rff_features}."
            )
        self._num_rff_features = num_rff_features

        if not isinstance(self.optimizer.optimizer, tf_keras.optimizers.Optimizer):
            raise ValueError(
                f"Optimizer for `DeepGaussianProcess` must be an instance of a "
                f"`tf.optimizers.Optimizer` or `tf.keras.optimizers.Optimizer`, "
                f"received {type(self.optimizer.optimizer)} instead."
            )

        if not isinstance(
            self.optimizer.optimizer.lr, tf_keras.optimizers.schedules.LearningRateSchedule
        ):
            self.original_lr = self.optimizer.optimizer.lr.numpy()

        epochs = 400

        def scheduler(epoch: int, lr: float) -> float:
            if epoch == epochs // 2:
                return lr * 0.1
            else:
                return lr

        if not self.optimizer.fit_args:
            self.optimizer.fit_args = {
                "verbose": 0,
                "epochs": epochs,
                "batch_size": 1000,
                "callbacks": [tf_keras.callbacks.LearningRateScheduler(scheduler)],
            }

        if self.optimizer.metrics is None:
            self.optimizer.metrics = ["mse"]

        self._model_gpflux = model
        # inputs and targets need to be redone with a float64 dtype to avoid setting the keras
        # backend to float64, this is likely to be fixed in GPflux, see issue:
        # https://github.com/secondmind-labs/GPflux/issues/76
        self._model_gpflux.inputs = tf_keras.Input(
            tuple(self._model_gpflux.inputs.shape[:-1]),
            name=self._model_gpflux.inputs.name,
            dtype=tf.float64,
        )
        self._model_gpflux.targets = tf_keras.Input(
            tuple(self._model_gpflux.targets.shape[:-1]),
            name=self._model_gpflux.targets.name,
            dtype=tf.float64,
        )
        self._model_keras = model.as_training_model()
        self._model_keras.compile(
            optimizer=self.optimizer.optimizer, metrics=self.optimizer.metrics, **compile_args
        )
        self._absolute_epochs = 0
        self._continuous_optimisation = continuous_optimisation

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()

        # when using a model closure, store the model parameters, not the model itself
        if self._model_closure is not None:
            state["_model_gpflux"] = gpflow.utilities.parameter_dict(self._model_gpflux)
            state["_model_keras"] = gpflow.utilities.parameter_dict(self._model_keras)

        # use to_json and get_weights to save any optimizer fit_arg callback models
        callbacks: list[Callback] = self._optimizer.fit_args.get("callbacks", [])
        callback: Callback
        saved_models: list[KerasOptimizer] = []
        tensorboard_writers: list[dict[str, Any]] = []
        try:
            for callback in callbacks:
                # serialize the callback models before pickling the optimizer
                saved_models.append(callback.model)
                if callback.model is self._model_keras:
                    # no need to serialize the main model, just use a special value instead
                    callback.model = ...
                elif callback.model:
                    callback.model = (callback.model.to_json(), callback.model.get_weights())
                # don't pickle tensorboard writers either; they'll be recreated when needed
                if isinstance(callback, tf_keras.callbacks.TensorBoard):
                    tensorboard_writers.append(callback._writers)
                    callback._writers = {}
            state["_optimizer"] = dill.dumps(state["_optimizer"])
        except Exception as e:
            raise NotImplementedError(
                "Failed to copy DeepGaussianProcess optimizer due to unsupported callbacks."
            ) from e
        finally:
            # revert original state, even if the pickling failed
            for callback, model in zip(self._optimizer.fit_args.get("callbacks", []), saved_models):
                callback.model = model
            for callback, writers in zip(
                (cb for cb in callbacks if isinstance(cb, tf_keras.callbacks.TensorBoard)),
                tensorboard_writers,
            ):
                callback._writers = writers

        # do the same thing for the history callback
        if self._model_keras.history:
            history_model = self._model_keras.history.model
            try:
                if history_model is self._model_keras:
                    # no need to serialize the main model, just use a special value instead
                    self._model_keras.history.model = ...
                elif history_model:
                    self._model_keras.history.model = (
                        history_model.to_json(),
                        history_model.get_weights(),
                    )
                state["_history"] = dill.dumps(self._model_keras.history)
            finally:
                self._model_keras.history.model = history_model

        # don't try to serialize any other copies of the history callback
        if isinstance(state.get("_last_optimization_result"), tf_keras.callbacks.History):
            state["_last_optimization_result"] = ...

        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.__dict__.update(state)

        # regenerate the models using the model closure
        if self._model_closure is not None:
            dgp: DeepGP = state["_model_closure"]()
            self._model_gpflux = dgp
            # inputs and targets need to be redone with a float64 dtype to avoid setting the keras
            # backend to float64, this is likely to be fixed in GPflux, see issue:
            # https://github.com/secondmind-labs/GPflux/issues/76
            self._model_gpflux.inputs = tf_keras.Input(
                tuple(self._model_gpflux.inputs.shape[:-1]),
                name=self._model_gpflux.inputs.name,
                dtype=tf.float64,
            )
            self._model_gpflux.targets = tf_keras.Input(
                tuple(self._model_gpflux.targets.shape[:-1]),
                name=self._model_gpflux.targets.name,
                dtype=tf.float64,
            )
            self._model_keras = dgp.as_training_model()

        # unpickle the optimizer, and restore all the callback models
        self._optimizer = dill.loads(self._optimizer)
        for callback in self._optimizer.fit_args.get("callbacks", []):
            if callback.model is ...:
                callback.set_model(self._model_keras)
            elif callback.model:
                model_json, weights = callback.model
                model = tf_keras.models.model_from_json(model_json)
                model.set_weights(weights)
                callback.set_model(model)

        # recompile the model
        self._model_keras.compile(self._optimizer.optimizer)

        # assign the model parameters
        if self._model_closure is not None:
            gpflow.utilities.multiple_assign(self._model_gpflux, state["_model_gpflux"])
            gpflow.utilities.multiple_assign(self._model_keras, state["_model_keras"])

        # restore the history (including any model it contains)
        if "_history" in state:
            self._model_keras.history = dill.loads(state["_history"])
            if self._model_keras.history.model is ...:
                self._model_keras.history.set_model(self._model_keras)
            elif self._model_keras.history.model:
                model_json, weights = self._model_keras.history.model
                model = tf_keras.models.model_from_json(model_json)
                model.set_weights(weights)
                self._model_keras.history.set_model(model)

        # recover optimization result if necessary (and possible)
        if state.get("_last_optimization_result") is ...:
            self._last_optimization_result = getattr(self._model_keras, "history")

    def __repr__(self) -> str:
        """"""
        return f"DeepGaussianProcess({self.model_gpflux!r}, {self.optimizer.optimizer!r})"

    @property
    def model_gpflux(self) -> DeepGP:
        return self._model_gpflux

    @property
    def model_keras(self) -> tf_keras.Model:
        return self._model_keras

    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        trajectory = self.trajectory_sampler().get_trajectory()
        expanded_query_points = tf.expand_dims(query_points, -2)  # [N, 1, D]
        tiled_query_points = tf.tile(expanded_query_points, [1, num_samples, 1])  # [N, S, D]
        return tf.transpose(trajectory(tiled_query_points), [1, 0, 2])  # [S, N, L]

    def reparam_sampler(self, num_samples: int) -> ReparametrizationSampler[GPfluxPredictor]:
        """
        Return a reparametrization sampler for a :class:`DeepGaussianProcess` model.

        :param num_samples: The number of samples to obtain.
        :return: The reparametrization sampler.
        """
        return DeepGaussianProcessReparamSampler(num_samples, self)

    def trajectory_sampler(self) -> TrajectorySampler[GPfluxPredictor]:
        """
        Return a trajectory sampler. For :class:`DeepGaussianProcess`, we build
        trajectories using the GPflux default sampler.

        :return: The trajectory sampler.
        """
        return DeepGaussianProcessDecoupledTrajectorySampler(self, self._num_rff_features)

    def update_encoded(self, dataset: Dataset) -> None:
        inputs = dataset.query_points
        new_num_data = inputs.shape[0]
        self.model_gpflux.num_data = new_num_data

        # Update num_data for each layer, as well as make sure dataset shapes are ok
        for i, layer in enumerate(self.model_gpflux.f_layers):
            if hasattr(layer, "num_data"):
                layer.num_data = new_num_data

            if isinstance(layer, LatentVariableLayer):
                inputs = layer(inputs)
                continue

            if isinstance(layer.inducing_variable, InducingPoints):
                inducing_variable = layer.inducing_variable
            else:
                inducing_variable = layer.inducing_variable.inducing_variable

            if inputs.shape[-1] != inducing_variable.Z.shape[-1]:
                raise ValueError(
                    f"Shape {inputs.shape} of input to layer {layer} is incompatible with shape"
                    f" {inducing_variable.Z.shape} of that layer. Trailing dimensions must match."
                )

            if (
                i == len(self.model_gpflux.f_layers) - 1
                and dataset.observations.shape[-1] != layer.q_mu.shape[-1]
            ):
                raise ValueError(
                    f"Shape {dataset.observations.shape} of new observations is incompatible"
                    f" with shape {layer.q_mu.shape} of existing observations. Trailing"
                    f" dimensions must match."
                )

            inputs = layer(inputs)

    def optimize_encoded(self, dataset: Dataset) -> tf_keras.callbacks.History:
        """
        Optimize the model with the specified `dataset`.
        :param dataset: The data with which to optimize the `model`.
        """
        fit_args = dict(self.optimizer.fit_args)

        # Tell optimizer how many epochs have been used before: the optimizer will "continue"
        # optimization across multiple BO iterations rather than start fresh at each iteration.
        # This allows us to monitor training across iterations.

        if "epochs" in fit_args:
            fit_args["epochs"] = fit_args["epochs"] + self._absolute_epochs

        hist = self.model_keras.fit(
            {"inputs": dataset.query_points, "targets": dataset.observations},
            **fit_args,
            initial_epoch=self._absolute_epochs,
        )

        if self._continuous_optimisation:
            self._absolute_epochs = self._absolute_epochs + len(hist.history["loss"])

        # Reset lr in case there was an lr schedule: a schedule will have changed the learning
        # rate, so that the next time we call `optimize` the starting learning rate would be
        # different. Therefore, we make sure the learning rate is set back to its initial value.
        # However, this is not needed for `LearningRateSchedule` instances.
        if not isinstance(
            self.optimizer.optimizer.lr, tf_keras.optimizers.schedules.LearningRateSchedule
        ):
            self.optimizer.optimizer.lr.assign(self.original_lr)

        return hist

    def log(self, dataset: Optional[Dataset] = None) -> None:
        """
        Log model training information at a given optimization step to the Tensorboard.
        We log a few summary statistics of losses, layer KL divergences and metrics (as provided in
        ``optimizer``): ``final`` value at the end of the training, ``diff`` value as a difference
        between inital and final epoch. We also log epoch statistics, but as histograms, rather
        than time series. We also log several training data based metrics, such as root mean square
        error between predictions and observations and several others.

        For custom logs user will need to subclass the model and overwrite this method.

        :param dataset: Optional data that can be used to log additional data-based model summaries.
        """
        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                logging.scalar("epochs/num_epochs", len(self.model_keras.history.epoch))
                for idx, layer in enumerate(self.model_gpflux.f_layers):
                    write_summary_kernel_parameters(layer.kernel, prefix=f"layer[{idx}]/")
                write_summary_likelihood_parameters(self.model_gpflux.likelihood_layer.likelihood)
                for k, v in self.model_keras.history.history.items():
                    logging.histogram(f"{k}/epoch", lambda: v)
                    logging.scalar(f"{k}/final", lambda: v[-1])
                    logging.scalar(f"{k}/diff", lambda: v[0] - v[-1])
                if dataset:
                    write_summary_data_based_metrics(
                        dataset=dataset, model=self, prefix="training_"
                    )



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains acquisition function builders and acquisition functions for Bayesian active
learning.
"""

from __future__ import annotations

import math
from typing import Optional, Sequence, Union

import tensorflow as tf
import tensorflow_probability as tfp

from ...data import Dataset
from ...models import ProbabilisticModel
from ...models.interfaces import FastUpdateModel, SupportsPredictJoint
from ...types import TensorType
from ...utils import DEFAULTS
from ..interface import AcquisitionFunction, AcquisitionFunctionClass, SingleModelAcquisitionBuilder


class PredictiveVariance(SingleModelAcquisitionBuilder[SupportsPredictJoint]):
    """
    Builder for the determinant of the predictive covariance matrix over the batch points.
    For a batch of size 1 it is the same as maximizing the predictive variance.
    """

    def __init__(self, jitter: float = DEFAULTS.JITTER) -> None:
        """
        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
        """
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""
        return f"PredictiveVariance(jitter={self._jitter!r})"

    def prepare_acquisition_function(
        self,
        model: SupportsPredictJoint,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.

        :return: The determinant of the predictive function.
        """
        if not isinstance(model, SupportsPredictJoint):
            raise NotImplementedError(
                f"PredictiveVariance only works with models that support "
                f"predict_joint; received {model!r}"
            )

        return predictive_variance(model, self._jitter)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: SupportsPredictJoint,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        return function  # no need to update anything


def predictive_variance(model: SupportsPredictJoint, jitter: float) -> AcquisitionFunction:
    """
    The predictive variance acquisition function for active learning, based on
    the determinant of the covariance (see :cite:`MacKay1992` for details).
    Note that the model needs to supply covariance of the joint marginal distribution,
    which can be expensive to compute.

    :param model: The model of the objective function.
    :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
            the covariance matrix.
    """

    @tf.function
    def acquisition(x: TensorType) -> TensorType:
        try:
            _, covariance = model.predict_joint(x)
        except NotImplementedError:
            raise ValueError(
                """
                PredictiveVariance only supports models with a predict_joint method.
                """
            )
        return tf.exp(tf.linalg.logdet(covariance + jitter))

    return acquisition


class ExpectedFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builder for the Expected feasibility acquisition function for identifying a failure or
    feasibility region. It implements two related sampling strategies called *bichon* criterion
    (:cite:`bichon2008efficient`) and *ranjan* criterion (:cite:`ranjan2008sequential`). The goal
    of both criteria is to sample points with a mean close to the threshold and a high variance.
    """

    def __init__(self, threshold: float, alpha: float = 1, delta: int = 1) -> None:
        """
        :param threshold: The failure or feasibility threshold.
        :param alpha: The parameter which determines the neighbourhood around the estimated contour
            line as a percentage of the posterior variance in which to allocate new points. Defaults
            to value of 1.
        :param delta: The parameter identifying which criterion is used, *bichon* for value of 1
            (default) and *ranjan* for value of 2.
        :raise ValueError (or InvalidArgumentError): If arguments are not a scalar, or `alpha` is
            not positive, or `delta` is not 1 or 2.
        """
        tf.debugging.assert_scalar(threshold)
        tf.debugging.assert_scalar(alpha)
        tf.debugging.assert_positive(alpha, message="Parameter alpha must be positive.")
        tf.debugging.assert_scalar(delta)
        tf.debugging.Assert(delta in [1, 2], [delta])

        self._threshold = threshold
        self._alpha = alpha
        self._delta = delta

    def __repr__(self) -> str:
        """"""
        return (
            f"ExpectedFeasibility(threshold={self._threshold!r}, alpha={self._alpha!r},"
            f" delta={self._delta!r})"
        )

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.
        :return: The expected feasibility function. This function will raise
            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
            greater than one.
        """
        return bichon_ranjan_criterion(model, self._threshold, self._alpha, self._delta)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        return function  # no need to update anything


def bichon_ranjan_criterion(
    model: ProbabilisticModel,
    threshold: float,
    alpha: float,
    delta: int,
) -> AcquisitionFunction:
    r"""
    Return the *bichon* criterion (:cite:`bichon2008efficient`) and *ranjan* criterion
    (:cite:`ranjan2008sequential`) used in Expected feasibility acquisition function for active
    learning of failure or feasibility regions.

    The problem of identifying a failure or feasibility region of a function :math:`f` can be
    formalized as estimating the excursion set, :math:`\Gamma^* = \{ x \in X: f(x) \ge T\}`, or
    estimating the contour line, :math:`C^* = \{ x \in X: f(x) = T\}`, for some threshold :math:`T`
    (see :cite:`bect2012sequential` for more details).

    It turns out that probabilistic models can be used as classifiers for identifying where
    excursion probability is larger than 1/2 and this idea is used to build many sequential
    sampling strategies. We follow :cite:`bect2012sequential` and use a formulation which provides
    a common expression for these two criteria:

    .. math:: \mathbb{E}[\max(0, (\alpha s(x))^\delta - |T - m(x)|^\delta)]

    Here :math:`m(x)` and :math:`s(x)` are the mean and standard deviation of the predictive
    posterior of a probabilistic model. *Bichon* criterion is obtained when :math:`\delta = 1` while
    *ranjan* criterion is obtained when :math:`\delta = 2`. :math:`\alpha>0` is another parameter
    that acts as a percentage of standard deviation of the posterior around the current boundary
    estimate where we want to sample. The goal is to sample a point with a mean close to the
    threshold :math:`T` and a high variance, so that the positive difference in the equation above
    is as large as possible.

    Note that only batches of size 1 are allowed.

    :param model: The probabilistic model of the objective function.
    :param threshold: The failure or feasibility threshold.
    :param alpha: The parameter which determines the neighbourhood around the estimated contour
        line as a percentage of the posterior variance in which to allocate new points.
    :param delta: The parameter identifying which criterion is used, *bichon* for value of 1
        and *ranjan* for value of 2.
    """

    @tf.function
    def acquisition(x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )

        mean, variance = model.predict(tf.squeeze(x, -2))
        stdev = tf.sqrt(variance)
        t = (threshold - mean) / stdev
        t_plus = t + alpha
        t_minus = t - alpha
        normal = tfp.distributions.Normal(tf.constant(0, x.dtype), tf.constant(1, x.dtype))

        if delta == 1:
            G = (
                alpha * (normal.cdf(t_plus) - normal.cdf(t_minus))
                - t * (2 * normal.cdf(t) - normal.cdf(t_plus) - normal.cdf(t_minus))
                - (2 * normal.prob(t) - normal.prob(t_plus) - normal.prob(t_minus))
            )
            tf.debugging.check_numerics(G, "NaN or Inf values encountered in criterion")
            criterion = G * stdev
        elif delta == 2:
            G = (
                (alpha**2 - 1 - t**2) * (normal.cdf(t_plus) - normal.cdf(t_minus))
                - 2 * t * (normal.prob(t_plus) - normal.prob(t_minus))
                + t_plus * normal.prob(t_plus)
                - t_minus * normal.prob(t_minus)
            )
            tf.debugging.check_numerics(G, "NaN or Inf values encountered in criterion")
            criterion = G * variance

        return criterion

    return acquisition


class IntegratedVarianceReduction(SingleModelAcquisitionBuilder[FastUpdateModel]):
    """
    Builder for the reduction of the integral of the predicted variance over the search
    space given a batch of query points.
    """

    def __init__(
        self,
        integration_points: TensorType,
        threshold: Optional[Union[float, Sequence[float], TensorType]] = None,
    ) -> None:
        """
        :param integration_points: set of points to integrate the prediction variance over.
        :param threshold: either None, a float or a sequence of 1 or 2 float values.
        """
        self._integration_points = integration_points
        self._threshold = threshold

    def __repr__(self) -> str:
        """"""
        return f"IntegratedVarianceReduction(threshold={self._threshold!r})"

    def prepare_acquisition_function(
        self,
        model: FastUpdateModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.

        :return: The integral of the predictive variance.
        """
        if not isinstance(model, FastUpdateModel):
            raise NotImplementedError(
                f"PredictiveVariance only works with FastUpdateModel models; received {model!r}"
            )

        return integrated_variance_reduction(model, self._integration_points, self._threshold)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: FastUpdateModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        return function  # no need to update anything


class integrated_variance_reduction(AcquisitionFunctionClass):
    r"""
    The reduction of the (weighted) average of the predicted variance over the integration points
    (a.k.a. Integrated Means Square Error or IMSE criterion).
    See :cite:`Picheny2010` for details.

    The criterion (to maximise) writes as:

        .. math:: \int_x (v_{old}(x) - v_{new}(x)) * weights(x),

    where :math:`v_{old}(x)` is the predictive variance of the model at :math:`x`, and
    :math:`v_{new}(x)` is the updated predictive variance, given that the GP is further
    conditioned on the query points.

    Note that since :math:`v_{old}(x)` is constant w.r.t. the query points, this function
    only returns :math:`-\int_x v_{new}(x) * weights(x)`.

    If no threshold is provided, the goal is to learn a globally accurate model, and
    the predictive variance (:math:`v_{new}`) is used. Otherwise, learning is 'targeted'
    towards regions where the GP is close to particular values, and the variance is weighted
    by the posterior GP pdf evaluated at the threshold T (if a single value is given) or by the
    probability that the GP posterior belongs to the interval between the 2 thresholds T1 and T2
    (note the slightly different parametrisation compared to :cite:`Picheny2010` in that case).

    This criterion allows batch size > 1. Note that the computational cost grows cubically with
    the batch size.

    This criterion requires a method (conditional_predict_f) to compute the new predictive variance
    given that query points are added to the data.
    """

    def __init__(
        self,
        model: FastUpdateModel,
        integration_points: TensorType,
        threshold: Optional[Union[float, Sequence[float], TensorType]] = None,
    ):
        """
        :param model: The model of the objective function.
        :param integration_points: Points over which to integrate the objective prediction variance.
        :param threshold: Either None, a float or a sequence of 1 or 2 float values.
            See class docs for details.
        :raise ValueError (or InvalidArgumentError): If ``threshold`` has more than 2 values.
        """
        self._model = model

        tf.debugging.assert_equal(
            len(tf.shape(integration_points)),
            2,
            message="integration_points must be of shape [N, D]",
        )

        tf.debugging.assert_positive(
            tf.shape(integration_points)[0],
            message="integration_points should contain at least one point",
        )

        self._integration_points = integration_points

        if threshold is None:
            self._weights = tf.constant(1.0, integration_points.dtype)

        else:
            if isinstance(threshold, float):
                t_threshold = tf.constant([threshold], integration_points.dtype)
            else:
                t_threshold = tf.cast(threshold, integration_points.dtype)

                tf.debugging.assert_rank(
                    t_threshold,
                    1,
                    message=f"threshold should be a float, a sequence "
                    f"or a rank 1 tensor, received {tf.shape(t_threshold)}",
                )
                tf.debugging.assert_less_equal(
                    tf.size(t_threshold),
                    2,
                    message=f"threshold should have one or two values,"
                    f" received {tf.size(t_threshold)}",
                )
                tf.debugging.assert_greater_equal(
                    tf.size(t_threshold),
                    1,
                    message=f"threshold should have one or two values,"
                    f" received {tf.size(t_threshold)}",
                )
                if tf.size(t_threshold) > 1:
                    tf.debugging.assert_greater_equal(
                        t_threshold[1],
                        t_threshold[0],
                        message=f"threshold values should be in increasing order,"
                        f" received {t_threshold}",
                    )

            if tf.size(t_threshold) == 1:
                mean_old, var_old = self._model.predict(query_points=integration_points)
                distr = tfp.distributions.Normal(mean_old, tf.sqrt(var_old))
                self._weights = distr.prob(t_threshold[0])
            else:
                mean_old, var_old = self._model.predict(query_points=integration_points)
                distr = tfp.distributions.Normal(mean_old, tf.sqrt(var_old))
                self._weights = distr.cdf(t_threshold[1]) - distr.cdf(t_threshold[0])

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        additional_data = Dataset(x, tf.ones_like(x[..., 0:1]))

        _, variance = self._model.conditional_predict_f(
            query_points=self._integration_points, additional_data=additional_data
        )

        return -tf.reduce_mean(variance * self._weights, axis=-2)


class BayesianActiveLearningByDisagreement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
    """
    Builder for the *Bayesian Active Learning By Disagreement* acquisition function defined in
    :cite:`houlsby2011bayesian`.
    """

    def __init__(self, jitter: float = DEFAULTS.JITTER) -> None:
        """
        :param jitter: The size of the jitter to avoid numerical problem caused by the
                log operation if variance is close to zero.
        """
        self._jitter = jitter

    def __repr__(self) -> str:
        """"""
        return f"BayesianActiveLearningByDisagreement(jitter={self._jitter!r})"

    def prepare_acquisition_function(
        self,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param model: The model.
        :param dataset: Unused.

        :return: The determinant of the predictive function.
        """

        return bayesian_active_learning_by_disagreement(model, self._jitter)

    def update_acquisition_function(
        self,
        function: AcquisitionFunction,
        model: ProbabilisticModel,
        dataset: Optional[Dataset] = None,
    ) -> AcquisitionFunction:
        """
        :param function: The acquisition function to update.
        :param model: The model.
        :param dataset: Unused.
        """
        return function  # no need to update anything


class bayesian_active_learning_by_disagreement(AcquisitionFunctionClass):
    def __init__(self, model: ProbabilisticModel, jitter: float):
        r"""
        The Bayesian active learning by disagrement acquisition function computes
        the information gain of the predictive entropy :cite:`houlsby2011bayesian`.
        the acquisiton function is calculated by:

        .. math::
            \mathrm{h}\left(\Phi\left(\frac{\mu_{\boldsymbol{x}, \mathcal{D}}}
            {\sqrt{\sigma_{\boldsymbol{x}, \mathcal{D}}^{2}+1}}\right)\right)
            -\frac{C \exp \left(-\frac{\mu_{\boldsymbol{x}, \mathcal{D}}^{2}}
            {2\left(\sigma_{\boldsymbol{w}, \mathcal{D}}^{+C^{2}}\right)}\right)}
            {\sqrt{\sigma_{\boldsymbol{x}, \mathcal{D}}^{2}+C^{2}}}

        Here :math:`\mathrm{h}(p)` is defined as:

        .. math::
            \mathrm{h}(p)=-p \log p-(1-p) \log (1-p)

        This acquisition function is intended to use for Binary Gaussian Process Classification
        model with Bernoulli likelihood. It is designed for VGP but other Gaussian approximation
        of the posterior can be used. SVGP for instance, or some other model that is not currently
        supported by Trieste. Integrating over nuisance parameters is currently not
        supported (see equation 6 of the paper).

        :param model: The model of the objective function.
        :param jitter: The size of the jitter to avoid numerical problem caused by the
                log operation if variance is close to zero.
        :return: The Bayesian Active Learning By Disagreement acquisition function.
        """
        tf.debugging.assert_positive(jitter, message="Jitter must be positive.")

        self._model = model
        self._jitter = jitter

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        tf.debugging.assert_shapes(
            [(x, [..., 1, None])],
            message="This acquisition function only supports batch sizes of one.",
        )
        mean, variance = self._model.predict(tf.squeeze(x, -2))
        variance = tf.maximum(variance, self._jitter)

        normal = tfp.distributions.Normal(tf.constant(0, mean.dtype), tf.constant(1, mean.dtype))
        p = normal.cdf((mean / tf.sqrt(variance + 1)))

        C2 = (math.pi * tf.math.log(tf.constant(2, mean.dtype))) / 2
        Ef = (tf.sqrt(C2) / tf.sqrt(variance + C2)) * tf.exp(-(mean**2) / (2 * (variance + C2)))

        return -p * tf.math.log(p + self._jitter) - (1 - p) * tf.math.log(1 - p + self._jitter) - Ef



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC
from typing import Callable, cast

import gpflow.kernels
import tensorflow as tf
from gpflow.inducing_variables import InducingPoints
from gpflux.layers import GPLayer, LatentVariableLayer
from gpflux.layers.basis_functions.fourier_features import RandomFourierFeaturesCosine
from gpflux.math import compute_A_inv_b
from gpflux.models import DeepGP

from ...types import TensorType
from ...utils import DEFAULTS, flatten_leading_dims
from ..interfaces import (
    ReparametrizationSampler,
    TrajectoryFunction,
    TrajectoryFunctionClass,
    TrajectorySampler,
)
from .interface import GPfluxPredictor


class DeepGaussianProcessReparamSampler(ReparametrizationSampler[GPfluxPredictor]):
    r"""
    This sampler employs the *reparameterization trick* to approximate samples from a
    :class:`GPfluxPredictor`\ 's predictive distribution, when the :class:`GPfluxPredictor` has
    an underlying :class:`~gpflux.models.DeepGP`.
    """

    def __init__(self, sample_size: int, model: GPfluxPredictor):
        """
        :param sample_size: The number of samples for each batch of points. Must be positive.
        :param model: The model to sample from.
        :raise ValueError (or InvalidArgumentError): If ``sample_size`` is not positive, if the
            model is not a :class:`GPfluxPredictor`, of if its underlying ``model_gpflux`` is not a
            :class:`~gpflux.models.DeepGP`.
        """
        if not isinstance(model, GPfluxPredictor):
            raise ValueError(
                f"Model must be a gpflux.interface.GPfluxPredictor, received {type(model)}"
            )

        super().__init__(sample_size, model)

        if not isinstance(self._model_gpflux, DeepGP):
            raise ValueError(
                f"GPflux model must be a gpflux.models.DeepGP, received {type(self._model_gpflux)}"
            )

        # Each element of _eps_list is essentially a lazy constant. It is declared and assigned an
        # empty tensor here, and populated on the first call to sample
        self._eps_list = [
            tf.Variable(
                tf.ones([sample_size, 0], dtype=self._model_gpflux.targets.dtype),
                shape=[sample_size, None],
            )
            for _ in range(len(self._model_gpflux.f_layers))
        ]
        self._encode = lambda x: model.encode(x)

    @property
    def _model_gpflux(self) -> tf.Module:
        return self._model.model_gpflux

    def sample(self, at: TensorType, *, jitter: float = DEFAULTS.JITTER) -> TensorType:
        """
        Return approximate samples from the `model` specified at :meth:`__init__`. Multiple calls to
        :meth:`sample`, for any given :class:`DeepGaussianProcessReparamSampler` and ``at``, will
        produce the exact same samples. Calls to :meth:`sample` on *different*
        :class:`DeepGaussianProcessReparamSampler` instances will produce different samples.

        :param at: Where to sample the predictive distribution, with shape `[..., 1, D]`, for points
            of dimension `D`.
        :param jitter: The size of the jitter to use when stabilizing the Cholesky
            decomposition of the covariance matrix.
        :return: The samples, of shape `[..., S, 1, L]`, where `S` is the `sample_size` and `L` is
            the number of latent model dimensions.
        :raise ValueError (or InvalidArgumentError): If ``at`` has an invalid shape or ``jitter``
            is negative.
        """
        tf.debugging.assert_shapes([(at, [..., 1, None])])
        tf.debugging.assert_greater_equal(jitter, 0.0)

        samples = tf.repeat(
            self._encode(at[..., None, :, :]), self._sample_size, axis=-3
        )  # [..., S, 1, D]
        for i, layer in enumerate(self._model_gpflux.f_layers):
            if isinstance(layer, LatentVariableLayer):
                if not self._initialized:
                    self._eps_list[i].assign(layer.prior.sample([tf.shape(samples)[:-1]]))
                samples = layer.compositor([samples, self._eps_list[i]])
                continue

            mean, var = layer.predict(samples, full_cov=False, full_output_cov=False)
            var = var + jitter

            if not self._initialized:
                self._eps_list[i].assign(
                    tf.random.normal([self._sample_size, tf.shape(mean)[-1]], dtype=tf.float64)
                )  # [S, L]

            samples = mean + tf.sqrt(var) * self._eps_list[i][:, None, :]

        if not self._initialized:
            self._initialized.assign(True)

        return samples  # [..., S, 1, L]


class DeepGaussianProcessDecoupledTrajectorySampler(TrajectorySampler[GPfluxPredictor]):
    r"""
    This sampler employs decoupled sampling (see :cite:`wilson2020efficiently`) to build functions
    that approximate a trajectory sampled from an underlying deep Gaussian process model. In
    particular, this sampler provides trajectory functions for :class:`GPfluxPredictor`\s with
    underlying :class:`~gpflux.models.DeepGP` models by using a feature decomposition using both
    random Fourier features and canonical features centered at inducing point locations. This allows
    for cheap approximate trajectory samples, as opposed to exact trajectory sampling, which scales
    cubically in the number of query points.
    """

    def __init__(
        self,
        model: GPfluxPredictor,
        num_features: int = 1000,
    ):
        """
        :param model: The model to sample from.
        :param num_features: The number of random Fourier features to use.
        :raise ValueError (or InvalidArgumentError): If the model is not a :class:`GPfluxPredictor`,
            or its underlying ``model_gpflux`` is not a :class:`~gpflux.models.DeepGP`, or
            ``num_features`` is not positive.
        """
        if not isinstance(model, GPfluxPredictor):
            raise ValueError(
                f"Model must be a gpflux.interface.GPfluxPredictor, received {type(model)}"
            )
        if not isinstance(model.model_gpflux, DeepGP):
            raise ValueError(
                f"GPflux model must be a gpflux.models.DeepGP, received {type(model.model_gpflux)}"
            )

        super().__init__(model)
        tf.debugging.assert_positive(num_features)
        self._num_features = num_features

    def __repr__(self) -> str:
        """"""
        return f"""{self.__class__.__name__}(
        {self._model!r},
        {self._num_features!r})
        """

    def get_trajectory(self) -> TrajectoryFunction:
        """
        Generate an approximate function draw (trajectory) from the deep GP model.

        :return: A trajectory function representing an approximate trajectory from the deep GP,
            taking an input of shape `[N, B, D]` and returning shape `[N, B, L]`.
        """

        return dgp_feature_decomposition_trajectory(self._model, self._num_features)

    def update_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Efficiently update a :const:`TrajectoryFunction` to reflect an update in its underlying
        :class:`ProbabilisticModel` and resample accordingly.

        :param trajectory: The trajectory function to be updated and resampled.
        :return: The updated and resampled trajectory function.
        :raise InvalidArgumentError: If ``trajectory`` is not a
            :class:`dgp_feature_decomposition_trajectory`
        """

        tf.debugging.Assert(
            isinstance(trajectory, dgp_feature_decomposition_trajectory), [tf.constant([])]
        )

        cast(dgp_feature_decomposition_trajectory, trajectory).update()
        return trajectory

    def resample_trajectory(self, trajectory: TrajectoryFunction) -> TrajectoryFunction:
        """
        Efficiently resample a :const:`TrajectoryFunction` in-place to avoid function retracing
        with every new sample.

        :param trajectory: The trajectory function to be resampled.
        :return: The new resampled trajectory function.
        :raise InvalidArgumentError: If ``trajectory`` is not a
            :class:`dgp_feature_decomposition_trajectory`
        """
        tf.debugging.Assert(
            isinstance(trajectory, dgp_feature_decomposition_trajectory), [tf.constant([])]
        )
        cast(dgp_feature_decomposition_trajectory, trajectory).resample()
        return trajectory


class DeepGaussianProcessDecoupledLayer(ABC):
    """
    Layer that samples an approximate decoupled trajectory for a GPflux
    :class:`~gpflux.layers.GPLayer` using Matheron's rule (:cite:`wilson2020efficiently`). Note
    that the only multi-output kernel that is supported is a
    :class:`~gpflow.kernels.SharedIndependent` kernel.
    """

    def __init__(
        self,
        model: GPfluxPredictor,
        layer_number: int,
        num_features: int = 1000,
    ):
        """
        :param model: The model to sample from.
        :param layer_number: The index of the layer that we wish to sample from.
        :param num_features: The number of features to use in the random feature approximation.
        :raise ValueError (or InvalidArgumentError): If the layer is not a
            :class:`~gpflux.layers.GPLayer`, the layer's kernel is not supported, or if
            ``num_features`` is not positive.
        """
        self._model = model
        self._layer_number = layer_number
        layer = self._layer

        if not isinstance(layer, GPLayer):
            raise ValueError(
                f"Layers other than gpflux.layers.GPLayer are not currently supported, received"
                f"{type(layer)}"
            )

        if isinstance(
            layer.inducing_variable, gpflow.inducing_variables.SeparateIndependentInducingVariables
        ):
            raise ValueError(
                "SeparateIndependentInducingVariables are not currently supported for decoupled "
                "sampling."
            )

        tf.debugging.assert_positive(num_features)
        self._num_features = num_features

        self._kernel = layer.kernel

        self._feature_functions = ResampleableDecoupledDeepGaussianProcessFeatureFunctions(
            layer, num_features
        )

        self._weight_sampler = self._prepare_weight_sampler()

        self._initialized = tf.Variable(False)

        self._weights_sample = tf.Variable(
            tf.ones([0, 0, 0], dtype=tf.float64), shape=[None, None, None]
        )

        self._batch_size = tf.Variable(0, dtype=tf.int32)

    @property
    def _layer(self) -> GPLayer:
        return self._model.model_gpflux.f_layers[self._layer_number]

    def __call__(self, x: TensorType) -> TensorType:  # [N, B, D] -> [N, B, P]
        """
        Evaluate trajectory function for layer at input.

        :param x: Input location with shape `[N, B, D]`, where `N` is the number of points, `B` is
            the batch dimension, and `D` is the input dimensionality.
        :return: Trajectory for the layer evaluated at the input, with shape `[N, B, P]`, where `P`
            is the number of latent GPs in the layer.
        :raise InvalidArgumentError: If the provided batch size does not match with the layer's
            batch size.
        """
        if not self._initialized:
            self._batch_size.assign(tf.shape(x)[-2])
            self.resample()
            self._initialized.assign(True)

        tf.debugging.assert_equal(
            tf.shape(x)[-2],
            self._batch_size.value(),
            message=f"""
            This trajectory only supports batch sizes of {self._batch_size}.
            If you wish to change the batch size you must get a new trajectory
            by calling the get_trajectory method of the trajectory sampler.
            """,
        )

        flat_x, unflatten = flatten_leading_dims(x)
        flattened_feature_evaluations = self._feature_functions(
            flat_x
        )  # [P, N, L + M] or [N, L + M]
        if self._feature_functions.is_multioutput:
            flattened_feature_evaluations = tf.transpose(
                flattened_feature_evaluations, perm=[1, 2, 0]
            )
            feature_evaluations = unflatten(flattened_feature_evaluations)  # [N, B, L + M, P]
        else:
            feature_evaluations = unflatten(flattened_feature_evaluations)[
                ..., None
            ]  # [N, B, L + M, 1]

        return tf.reduce_sum(
            feature_evaluations * self._weights_sample, -2
        ) + self._layer.mean_function(
            x
        )  # [N, B, P]

    def resample(self) -> None:
        """
        Efficiently resample in-place without retracing.
        """
        self._weights_sample.assign(self._weight_sampler(self._batch_size))

    def update(self) -> None:
        """
        Efficiently update the trajectory with a new weight distribution and resample its weights.
        """
        self._feature_functions.resample()
        self._weight_sampler = self._prepare_weight_sampler()
        self.resample()

    def _prepare_weight_sampler(self) -> Callable[[int], TensorType]:  # [B] -> [B, L+M, P]
        """
        Prepare the sampler function that provides samples of the feature weights for both the
        RFF and canonical feature functions, i.e. we return a function that takes in a batch size
        `B` and returns `B` samples for the weights of each of the `L` RFF features and `M`
        canonical features for `P` outputs.
        """

        if isinstance(self._layer.inducing_variable, InducingPoints):
            inducing_points = self._layer.inducing_variable.Z  # [M, D]
        else:
            inducing_points = self._layer.inducing_variable.inducing_variable.Z  # [M, D]

        q_mu = self._layer.q_mu  # [M, P]
        q_sqrt = self._layer.q_sqrt  # [P, M, M]
        if self._feature_functions.is_multioutput:
            Kmm = self._kernel.K(
                inducing_points, inducing_points, full_output_cov=False
            )  # [P, M, M]
        else:
            Kmm = self._kernel.K(inducing_points, inducing_points)  # [M, M]
        Kmm += tf.eye(tf.shape(inducing_points)[0], dtype=Kmm.dtype) * DEFAULTS.JITTER
        whiten = self._layer.whiten
        M, P = tf.shape(q_mu)[0], tf.shape(q_mu)[1]

        tf.debugging.assert_shapes(
            [
                (inducing_points, ["M", "D"]),
                (q_mu, ["M", "P"]),
                (q_sqrt, ["P", "M", "M"]),
            ]
        )

        def weight_sampler(batch_size: int) -> TensorType:
            prior_weights = tf.random.normal(
                [batch_size, P, self._num_features, 1], dtype=tf.float64
            )  # [B, P, L, 1]

            u_noise_sample = tf.matmul(
                q_sqrt,  # [P, M, M]
                tf.random.normal([batch_size, P, M, 1], dtype=tf.float64),  # [B, P, M, 1]
            )  # [B, P, M, 1]
            u_sample = tf.linalg.matrix_transpose(q_mu)[..., None] + u_noise_sample  # [B, P, M, 1]

            if whiten:
                Luu = tf.linalg.cholesky(Kmm)  # [M, M] or [P, M, M]
                u_sample = tf.matmul(Luu, u_sample)  # [B, P, M, 1]

            phi_Z = self._feature_functions(inducing_points)[
                ..., : self._num_features
            ]  # [M, L] or [P, M, L]
            weight_space_prior_Z = phi_Z @ prior_weights  # [B, P, M, 1]

            diff = u_sample - weight_space_prior_Z  # [B, P, M, 1]
            v = compute_A_inv_b(Kmm, diff)  # [B, P, M, 1]

            return tf.transpose(
                tf.concat([prior_weights, v], axis=2)[..., 0], perm=[0, 2, 1]
            )  # [B, L + M, P]

        return weight_sampler


class ResampleableDecoupledDeepGaussianProcessFeatureFunctions(RandomFourierFeaturesCosine):
    """
    A wrapper around GPflux's random Fourier feature function that allows for efficient in-place
    updating when generating new decompositions. In addition to providing Fourier features,
    this class concatenates a layer's Fourier feature expansion with evaluations of the canonical
    basis functions.
    """

    def __init__(self, layer: GPLayer, n_components: int):
        """
        :param layer: The layer that will be approximated by the feature functions.
        :param n_components: The number of features.
        :raise ValueError: If the layer is not a :class:`~gpflux.layers.GPLayer`.
        """
        if not isinstance(layer, GPLayer):
            raise ValueError(
                f"ResampleableDecoupledDeepGaussianProcessFeatureFunctions currently only work with"
                f"gpflux.layers.GPLayer layers, received {type(layer)} instead"
            )

        self._kernel = layer.kernel

        self._n_components = n_components
        super().__init__(self._kernel, self._n_components, dtype=tf.float64)

        if isinstance(layer.inducing_variable, InducingPoints):
            inducing_points = layer.inducing_variable.Z
        else:
            inducing_points = layer.inducing_variable.inducing_variable.Z

        if self.is_multioutput:
            self._canonical_feature_functions = lambda x: tf.linalg.matrix_transpose(
                self._kernel.K(inducing_points, x, full_output_cov=False)
            )
        else:
            self._canonical_feature_functions = lambda x: tf.linalg.matrix_transpose(
                self._kernel.K(inducing_points, x)
            )

        dummy_X = inducing_points[0:1, :]

        self.__call__(dummy_X)
        self.b: TensorType = tf.Variable(self.b)
        self.W: TensorType = tf.Variable(self.W)

    def resample(self) -> None:
        """
        Resample weights and biases.
        """
        if not hasattr(self, "_bias_init"):
            self.b.assign(self._sample_bias(tf.shape(self.b), dtype=self._dtype))
            self.W.assign(self._sample_weights(tf.shape(self.W), dtype=self._dtype))
        else:
            self.b.assign(self._bias_init(tf.shape(self.b), dtype=self._dtype))
            self.W.assign(self._weights_init(tf.shape(self.W), dtype=self._dtype))

    def __call__(self, x: TensorType) -> TensorType:  # [N, D] -> [N, L + M] or [P, N, L + M]
        """
        Evaluate and combine prior basis functions and canonical basic functions at the input.
        """
        fourier_feature_eval = super().__call__(x)  # [N, L] or [P, N, L]
        canonical_feature_eval = self._canonical_feature_functions(x)  # [P, N, M] or [N, M]
        return tf.concat([fourier_feature_eval, canonical_feature_eval], axis=-1)  # [P, N, L + M]


class dgp_feature_decomposition_trajectory(TrajectoryFunctionClass):
    r"""
    An approximate sample from a deep Gaussian process's posterior, where the samples are
    represented as a finite weighted sum of features. This class essentially takes a list of
    :class:`DeepGaussianProcessDecoupledLayer`\s and iterates through them to sample, update and
    resample.
    """

    def __init__(self, model: GPfluxPredictor, num_features: int):
        """
        :param model: The model to sample from.
        :param num_features: The number of random Fourier features to use.
        """
        self._sampling_layers = [
            DeepGaussianProcessDecoupledLayer(model, i, num_features)
            for i in range(len(model.model_gpflux.f_layers))
        ]

        self._encode = lambda x: model.encode(x)

    @tf.function
    def __call__(self, x: TensorType) -> TensorType:
        """
        Call trajectory function by looping through layers.

        :param x: Input location with shape `[N, B, D]`, where `N` is the number of points, `B` is
            the batch dimension, and `D` is the input dimensionality.
        :return: Trajectory samples with shape `[N, B, L]`, where `L` is the number of outputs.
        """
        x = self._encode(x)
        for layer in self._sampling_layers:
            x = layer(x)
        return x

    def update(self) -> None:
        """Update the layers with new features and weights."""
        for layer in self._sampling_layers:
            layer.update()

    def resample(self) -> None:
        """Resample the layer weights."""
        for layer in self._sampling_layers:
            layer.resample()



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" This module contains logging utilities. """
from __future__ import annotations

import io
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, TypeVar, Union

import absl
import tensorflow as tf
from tensorflow.python.eager import context

from trieste.types import TensorType

if TYPE_CHECKING:
    import matplotlib


SummaryFilter = Callable[[str], bool]


def default_summary_filter(name: str) -> bool:
    """Default summary filter: omits any names that start with _."""
    return not (name.startswith("_") or "/_" in name)


_TENSORBOARD_WRITER: Optional[tf.summary.SummaryWriter] = None
_STEP_NUMBER: int = 0
_SUMMARY_FILTER: SummaryFilter = default_summary_filter


def set_tensorboard_writer(summary_writer: Optional[tf.summary.SummaryWriter]) -> None:
    """
    Set a :class:`~tf.summary.SummaryWriter` instance to use for logging
    to TensorBoard, or `None` to disable.

    :param summary_writer: optional summary writer instance.
    """
    global _TENSORBOARD_WRITER
    _TENSORBOARD_WRITER = summary_writer


def get_tensorboard_writer() -> Optional[tf.summary.SummaryWriter]:
    """
    Returns a :class:`~tf.summary.SummaryWriter` instance to use for logging
    to TensorBoard, or `None`.

    :return: optional summary writer instance.
    """
    return _TENSORBOARD_WRITER


@contextmanager
def tensorboard_writer(summary_writer: Optional[tf.summary.SummaryWriter]) -> Iterator[None]:
    """
    A context manager for setting or overriding a TensorBoard summary writer inside a code block.

    :param summary_writer: optional summary writer instance.
    """
    old_writer = get_tensorboard_writer()
    set_tensorboard_writer(summary_writer)
    yield
    set_tensorboard_writer(old_writer)


def set_step_number(new_step_number: int) -> None:
    """
    Set an optimization step number to use for logging purposes.

    :param new_step_number: new current step number
    :raise ValueError: if step_number < 0
    """
    global _STEP_NUMBER
    _STEP_NUMBER = new_step_number


def get_step_number() -> int:
    """
    Get the optimization step number used for logging purposes.

    :return: current step number.
    """
    return _STEP_NUMBER


@contextmanager
def step_number(new_step_number: int) -> Iterator[None]:
    """
    A context manager for setting or overriding the optimization step number inside a code block.

    :param new_step_number: new current step number
    """
    old_step_number = get_step_number()
    set_step_number(new_step_number)
    yield
    set_step_number(old_step_number)


def set_summary_filter(summary_filter: SummaryFilter) -> None:
    """
    Set a filter on summary names. The default is to only omit names that start with _.

    :param summary_filter: new summary filter
    """
    global _SUMMARY_FILTER
    _SUMMARY_FILTER = summary_filter


def get_summary_filter() -> SummaryFilter:
    """
    Get the current filter on summary names. The default is to only omit names that start with _.

    :return: current summary filter.
    """
    return _SUMMARY_FILTER


def get_current_name_scope() -> str:
    """Returns the full name scope. Copied from TF 2.5."""
    ctx = context.context()
    if ctx.executing_eagerly():
        return ctx.scope_name.rstrip("/")
    else:
        return tf.compat.v1.get_default_graph().get_name_scope()


def include_summary(name: str) -> bool:
    """
    Whether a summary name should be included.

    :param: full summary name (including name spaces)
    :return: whether the summary should be included.
    """
    full_name = get_current_name_scope() + "/" + name
    return _SUMMARY_FILTER(full_name)


T = TypeVar("T")


def evaluate_data(data: T | Callable[[], T]) -> T:
    """Return the passed in data, evaluating it if it's inside a closure."""
    return data() if callable(data) else data


def histogram(name: str, data: TensorType | Callable[[], TensorType], **kwargs: Any) -> bool:
    """
    Wrapper for tf.summary.histogram that first filters out unwanted summaries by name.
    Accepts either data or closures that only get evaluated when logged.
    """
    if include_summary(name):
        try:
            return tf.summary.histogram(name, evaluate_data(data), **kwargs)
        except Exception as e:
            tf.print(
                f"Failed to write tensorboard histogram summary '{name}':\n\n{e}",
                output_stream=absl.logging.INFO,
            )
    return False


def scalar(name: str, data: float | Callable[[], float], **kwargs: Any) -> bool:
    """
    Wrapper for tf.summary.scalar that first filters out unwanted summaries by name.
    Accepts either data or closures that only get evaluated when logged.
    """
    if include_summary(name):
        try:
            return tf.summary.scalar(name, evaluate_data(data), **kwargs)
        except Exception as e:
            tf.print(
                f"Failed to write tensorboard scalar summary '{name}':\n\n{e}",
                output_stream=absl.logging.INFO,
            )
    return False


def text(name: str, data: str | Callable[[], str], **kwargs: Any) -> bool:
    """
    Wrapper for tf.summary.text that first filters out unwanted summaries by name.
    Accepts either data or closures that only get evaluated when logged.
    """
    if include_summary(name):
        try:
            return tf.summary.text(name, evaluate_data(data), **kwargs)
        except Exception as e:
            tf.print(
                f"Failed to write tensorboard text summary '{name}':\n\n{e}",
                output_stream=absl.logging.INFO,
            )
    return False


def pyplot(
    name: str, figure: Union["matplotlib.figure.Figure", Callable[[], "matplotlib.figure.Figure"]]
) -> bool:
    """
    Utility function for passing a matplotlib figure to tf.summary.image.
    Accepts either data or closures that only get evaluated when logged.
    """
    if include_summary(name):
        try:
            figure = evaluate_data(figure)
            with io.BytesIO() as buffer:
                figure.savefig(buffer, dpi=150.0, format="png")
                buffer.seek(0)
                image = tf.image.decode_png(buffer.getvalue(), channels=4)
            image = tf.expand_dims(image, 0)
            return tf.summary.image(name, image)
        except Exception as e:
            tf.print(
                f"Failed to write tensorboard image summary '{name}':\n\n{e}",
                output_stream=absl.logging.INFO,
            )
    return False



# Copyright 2020 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The library root. See :mod:`~trieste.bayesian_optimizer` for the core optimizer, which requires
models (see :mod:`~trieste.models`), and data sets (see :mod:`~trieste.data`). The
:mod:`~trieste.acquisition` package provides a selection of acquisition algorithms and the
functionality to define your own. The :mod:`~trieste.ask_tell_optimization` package provides API
for Ask-Tell optimization and manual control of the optimization loop.
The :mod:`~trieste.objectives` package contains several popular objective functions,
useful for experimentation.
"""
from . import (
    acquisition,
    ask_tell_optimization,
    bayesian_optimizer,
    data,
    models,
    objectives,
    observer,
    space,
    types,
    utils,
)
from .version import __version__



# Copyright 2021 The Trieste Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Optional

import gpflow
import tensorflow as tf
from gpflow.models import GPModel
from gpflow.posteriors import BasePosterior, PrecomputeCacheType
from typing_extensions import Protocol, final

from ... import logging
from ...data import Dataset
from ...space import EncoderFunction
from ...types import TensorType
from ..interfaces import (
    EncodedProbabilisticModel,
    EncodedSupportsPredictJoint,
    EncodedSupportsPredictY,
    EncodedTrainableProbabilisticModel,
    HasReparamSampler,
    ReparametrizationSampler,
    SupportsGetKernel,
    SupportsGetObservationNoise,
    SupportsPredictJoint,
)
from ..optimizer import Optimizer
from ..utils import (
    write_summary_data_based_metrics,
    write_summary_kernel_parameters,
    write_summary_likelihood_parameters,
)
from .sampler import BatchReparametrizationSampler


class GPflowPredictor(
    EncodedSupportsPredictJoint,
    SupportsGetKernel,
    SupportsGetObservationNoise,
    EncodedSupportsPredictY,
    HasReparamSampler,
    EncodedTrainableProbabilisticModel,
    EncodedProbabilisticModel,
    ABC,
):
    """A trainable wrapper for a GPflow Gaussian process model."""

    def __init__(self, optimizer: Optimizer | None = None, encoder: EncoderFunction | None = None):
        """
        :param optimizer: The optimizer with which to train the model. Defaults to
            :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
        :param encoder: Optional encoder with which to transform query points before
            generating predictions.
        """
        if optimizer is None:
            optimizer = Optimizer(gpflow.optimizers.Scipy(), compile=True)

        self._optimizer = optimizer
        self._encoder = encoder
        self._posterior: Optional[BasePosterior] = None

    @property
    def encoder(self) -> EncoderFunction | None:
        return self._encoder

    @encoder.setter
    def encoder(self, encoder: EncoderFunction | None) -> None:
        self._encoder = encoder

    @property
    def optimizer(self) -> Optimizer:
        """The optimizer with which to train the model."""
        return self._optimizer

    def create_posterior_cache(self) -> None:
        """
        Create a posterior cache for fast sequential predictions.  Note that this must happen
        at initialisation and *after* we ensure the model data is variable. Furthermore,
        the cache must be updated whenever the underlying model is changed.
        """
        self._posterior = self.model.posterior(PrecomputeCacheType.VARIABLE)

    def _ensure_variable_model_data(self) -> None:
        """Ensure GPflow data, which is normally stored in Tensors, is instead stored in
        dynamically shaped Variables. Override this as required."""

    def __setstate__(self, state: dict[str, Any]) -> None:
        # when unpickling we may need to regenerate the posterior cache
        self.__dict__.update(state)
        self._ensure_variable_model_data()
        if self._posterior is not None:
            self.create_posterior_cache()

    def update_posterior_cache(self) -> None:
        """Update the posterior cache. This needs to be called whenever the underlying model
        is changed."""
        if self._posterior is not None:
            self._posterior.update_cache()

    @property
    @abstractmethod
    def model(self) -> GPModel:
        """The underlying GPflow model."""

    def predict_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        mean, cov = (self._posterior or self.model).predict_f(query_points)
        # posterior predict can return negative variance values [cf GPFlow issue #1813]
        if self._posterior is not None:
            cov = tf.clip_by_value(cov, 1e-12, cov.dtype.max)
        return mean, cov

    def predict_joint_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        mean, cov = (self._posterior or self.model).predict_f(query_points, full_cov=True)
        # posterior predict can return negative variance values [cf GPFlow issue #1813]
        if self._posterior is not None:
            cov = tf.linalg.set_diag(
                cov, tf.clip_by_value(tf.linalg.diag_part(cov), 1e-12, cov.dtype.max)
            )
        return mean, cov

    def sample_encoded(self, query_points: TensorType, num_samples: int) -> TensorType:
        return self.model.predict_f_samples(query_points, num_samples)

    def predict_y_encoded(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
        return self.model.predict_y(query_points)

    def get_kernel(self) -> gpflow.kernels.Kernel:
        """
        Return the kernel of the model.

        :return: The kernel.
        """
        return self.model.kernel

    def get_mean_function(self) -> gpflow.mean_functions.MeanFunction:
        """
        Return the mean function of the model.

        :return: The mean function.
        """
        return self.model.mean_function

    def get_observation_noise(self) -> TensorType:
        """
        Return the variance of observation noise for homoscedastic likelihoods.

        :return: The observation noise.
        :raise NotImplementedError: If the model does not have a homoscedastic likelihood.
        """
        try:
            noise_variance = self.model.likelihood.variance
        except AttributeError:
            raise NotImplementedError(f"Model {self!r} does not have scalar observation noise")

        return noise_variance

    def log(self, dataset: Optional[Dataset] = None) -> None:
        """
        Log model training information at a given optimization step to the Tensorboard.
        We log kernel and likelihood parameters. We also log several training data based metrics,
        such as root mean square error between predictions and observations and several others.

        :param dataset: Optional data that can be used to log additional data-based model summaries.
        """
        summary_writer = logging.get_tensorboard_writer()
        if summary_writer:
            with summary_writer.as_default(step=logging.get_step_number()):
                write_summary_kernel_parameters(self.get_kernel())
                write_summary_likelihood_parameters(self.model.likelihood)
                if dataset:
                    write_summary_data_based_metrics(
                        dataset=dataset, model=self, prefix="training_"
                    )

    def reparam_sampler(self, num_samples: int) -> ReparametrizationSampler[GPflowPredictor]:
        """
        Return a reparametrization sampler providing `num_samples` samples.

        :return: The reparametrization sampler.
        """
        return BatchReparametrizationSampler(num_samples, self)


class SupportsCovarianceBetweenPoints(SupportsPredictJoint, Protocol):
    """A probabilistic model that supports covariance_between_points."""

    @abstractmethod
    def covariance_between_points(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        r"""
        Compute the posterior covariance between sets of query points.

        .. math:: \Sigma_{12} = K_{12} - K_{x1}(K_{xx} + \sigma^2 I)^{-1}K_{x2}

        Note that query_points_2 must be a rank 2 tensor, but query_points_1 can
        have leading dimensions.

        :param query_points_1: Set of query points with shape [..., N, D]
        :param query_points_2: Sets of query points with shape [M, D]
        :return: Covariance matrix between the sets of query points with shape [..., L, N, M]
            (L being the number of latent GPs = number of output dimensions)
        """
        raise NotImplementedError


class EncodedSupportsCovarianceBetweenPoints(
    EncodedProbabilisticModel, SupportsCovarianceBetweenPoints
):
    @abstractmethod
    def covariance_between_points_encoded(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        """Implementation of covariance_between_points on encoded query points."""

    @final
    def covariance_between_points(
        self, query_points_1: TensorType, query_points_2: TensorType
    ) -> TensorType:
        return self.covariance_between_points_encoded(
            self.encode(query_points_1), self.encode(query_points_2)
        )
